#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT


class Error(Exception):

    """Base class for other exceptions."""

    def __init__(self, *message):
        self.__message = r' '.join([str(m) for m in message])
        super().__init__(self.__message)

    def __str__(self):
        return self.__message


__all__ = ['Error']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

from .configurable import Configurable
from .metavars import *
from .writer import *


class Column(Configurable):
    def __init__(self, var):
        super().__init__(var)
        self.variable = var
        self.struct = var.struct
        self.index = -1  # set by the struct
        self.name = var.name
        self.alignment = var.alignment
        self.type = var.type
        self.param_type = var.param_type
        self.pointer_type = var.pointer_type
        self.const_pointer_type = var.const_pointer_type
        self.default = var.default


__all__ = [r'Column']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

from io import StringIO

from . import cpp, log, utils
from .column import *
from .configurable import *
from .includes import *
from .metavars import *
from .schemas import *
from .type_list import *
from .variable import *
from .writer import *

NEWLINE = '\n'
TAB = '\t'
TAB_SPACES = '    '


class Struct(Configurable):
    __schema = Schema(
        {
            Optional(r'allocator', default=''): Stripped(str),
            Optional(r'annotations', default=[]): And(
                ValueOrArray(str, name=r'annotations'),
                Use(lambda x: utils.remove_duplicates([s.strip() for s in x if s.strip()])),
            ),
            Optional(r'attributes', default=[]): And(
                ValueOrArray(str, name=r'attributes'),
                Use(lambda x: utils.remove_duplicates([s.strip() for s in x if s.strip()])),
            ),
            Optional(r'brief', default=''): Stripped(str),
            Optional(r'copyable', default=True): bool,
            Optional(r'default_constructible', default=True): Or(bool, 'auto'),
            Optional(r'details', default=''): Stripped(str),
            Optional(r'epilogue', default=''): Stripped(str),
            Optional(r'footer', default=''): Stripped(str),
            Optional(r'header', default=''): Stripped(str),
            Optional(r'includes', default=dict): {object: object},
            Optional(r'movable', default=True): bool,
            Optional(r'prologue', default=''): Stripped(str),
            Optional(r'reverse_iterators', default=False): bool,
            Optional(r'static_variables', default=list): [object],
            Optional(r'swappable', default=True): bool,
            Optional(r'variables', default=list): [object],
        }
    )

    def __init__(self, cfg, name, vals):
        super().__init__(cfg)

        vals = Struct.__schema.validate(vals)
        self.__dict__.update(vals)

        self.name = name
        valid = cpp.is_valid_identifier(self.name)
        if not valid[0]:
            raise SchemaError(rf"name {valid[1]}", None)
        self.type = name

        if self.reverse_iterators:
            log.w(rf'{current_schema_context()}reverse_iterators: not currently implemented')
            self.reverse_iterators = False

        self.qualified_type = rf'{self.config.namespace}::{self.type}' if self.config.namespace else self.type
        self.qualified_name = self.qualified_type
        self.index = -1  # set by the config

        self.meta = MetaVars()
        self.meta.push(r'name', self.name)
        self.meta.push(r'type', self.type)
        self.meta.push(r'qualified_name', self.qualified_name)
        self.meta.push(r'qualified_type', self.qualified_type)
        self.meta.push(r'struct::name', self.name)
        self.meta.push(r'struct::type', self.type)
        self.meta.push(r'struct::qualified_name', self.qualified_name)
        self.meta.push(r'struct::qualified_type', self.qualified_type)
        self.meta.push(r'struct::scope', '')

        if not self.allocator:
            self.allocator = self.config.allocator
        if self.allocator == r'std::allocator':
            self.allocator = r'std::allocator<std::byte>'

        self.annotations = utils.remove_duplicates(self.config.all_structs.annotations + self.annotations)
        self.attributes = utils.remove_duplicates(self.config.all_structs.attributes + self.attributes)
        if r'SOAGEN_EMPTY_BASES' not in self.attributes:
            self.attributes.append(r'SOAGEN_EMPTY_BASES')
        if r'__declspec(empty_bases)' in self.attributes:
            self.attributes.remove(r'__declspec(empty_bases)')

        with SchemaContext('includes'):
            self.includes = Includes(cfg, self.includes)

        has_defaults = False
        for i in range(len(self.variables)):
            with SchemaContext(
                rf"variable '{self.variables[i]['name']}'" if r'name' in self.variables[i] else rf"variable [{i}]"
            ):
                v = Variable(self, self.variables[i])
                self.variables[i] = v
                if v.name == self.name:
                    raise SchemaError(rf"name: may not be the same as the struct", None)

                self.meta.push(rf'{v.name}::type', v.type)
                self.meta.push(rf'{v.name}::default', v.default if v.default else '{}')
                has_default = bool(v.default)
                if not has_default and has_defaults:
                    log.w(rf'{current_schema_context()}has no default but previous variables do')
                has_defaults = has_defaults or has_default

        self.columns = []
        index = 0
        for v in self.variables:
            v.index = index
            index += 1
            self.columns += v.columns

        index = 0
        for c in self.columns:
            c.index = index
            index += 1
        self.column_indices = ", ".join([str(col.index) for col in self.columns])

        self.prologue = rf'''
        {self.config.all_structs.prologue}

        {self.prologue}
        '''.strip()

        self.header = rf'''
        {self.config.all_structs.header}

        {self.header}
        '''.strip()

        self.footer = rf'''
        {self.config.all_structs.footer}

        {self.footer}
        '''.strip()

        self.epilogue = rf'''
        {self.config.all_structs.epilogue}

        {self.epilogue}
        '''.strip()

        if not self.brief:
            self.brief = name
        self.details += rf'''

        @remark Models the <a href="https://en.wikipedia.org/wiki/AoS_and_SoA">Structure-of-arrays</a>
                equivalent of: @code{{.cpp}}
        struct {self.name}
        {{'''
        for col in self.columns:
            self.details += f'\n{{% tab %}}{col.type} {col.name};'
        self.details += '''
        };
        @endcode
        '''

        static_vars = [v for v in self.config.all_structs.static_variables]
        for i in range(len(self.static_variables)):
            with SchemaContext(
                rf"static variable '{self.static_variables[i]['name']}'"
                if r'name' in self.static_variables[i]
                else rf"static variable [{i}]"
            ):
                static_vars.append(StaticVariable(self.config, self.static_variables[i]))
        static_vars = [(v, v.access) for v in static_vars]
        self.static_variables = {'public': [], 'protected': [], 'private': []}
        for v, access in static_vars:
            self.static_variables[access].append(v)
        assert len(self.static_variables) == 3

    def set_index(self, index):
        assert self.index == -1
        assert isinstance(index, int)
        assert index >= 0
        self.index = index
        self.meta.push('index', index)
        self.meta.push('struct::index', index)

    def write_forward_declarations(self, o: Writer):
        with MetaScope(self):
            o(rf'class {self.type};')

    def write_soagen_detail_specializations(self, o: Writer):
        with MetaScope(self):
            for col in self.columns:
                o(rf'SOAGEN_MAKE_NAMED_COLUMN({self.qualified_name}, {col.index}, {col.name});')
            max_length = 0
            for col in self.columns:
                max_length = max(len(col.name), max_length)
            with StringIO() as buf:
                buf.write('table_traits<\n')
                for i in range(len(self.columns)):
                    if i:
                        buf.write(f',\n')
                    col = self.columns[i]
                    buf.write(f'{o.indent_str*5}/* {col.name:>{max_length}} */ column_traits<{col.type}')
                    if col.alignment > 0:
                        buf.write(rf', soagen::max(size_t{{ {col.alignment} }}, alignof({col.type}))')
                    if col.param_type:
                        if not col.param_type:
                            buf.write(rf', alignof({col.type})')
                        buf.write(rf', {col.param_type}')
                    buf.write(rf'>')
                buf.write(rf'>')
                o(
                    rf'''
                template <>
                struct is_soa_<{self.qualified_name}> : std::true_type
                {{}};

                template <>
                struct table_traits_type_<{self.qualified_name}>
                {{
                    using type = {buf.getvalue()};
                }};

                template <>
                struct allocator_type_<{self.qualified_name}>
                {{
                    using type = {self.allocator};
                }};

                template <>
                struct table_type_<{self.qualified_name}>
                {{
                    using type = table<table_traits_type<{self.qualified_name}>, allocator_type<{self.qualified_name}>>;
                }};
                '''
                )

    def write_class_definition(self, o: Writer):
        with MetaScope(self):
            if self.prologue:
                o(
                    f'''
                {self.prologue}
                '''
                )

            def doxygen(s: str) -> str:
                nonlocal o
                if not o.doxygen:
                    return ''
                s = [x.rstrip() for x in s.split('\n')]
                popped_start = 0
                while len(s) and not s[0]:
                    s.pop(0)
                    popped_start += 1
                leading_spaces = None
                for i in range(len(s)):
                    if not s[i]:
                        continue
                    lstripped = s[i].lstrip()
                    if s[i] != lstripped:
                        leading_spaces = min(
                            len(s[i]) - len(lstripped), 999999 if leading_spaces is None else leading_spaces
                        )
                    s[i] = lstripped
                leading_spaces = ' ' * leading_spaces if leading_spaces is not None else ''
                s = f'{popped_start*NEWLINE}{leading_spaces}/// {rf"{NEWLINE}{leading_spaces}/// ".join(s)}'
                return s

            o(
                doxygen(
                    rf'''
                @brief {self.brief}

                @details {self.details}

                @note The code and documentation for this class were generated by soagen - https://marzer.github.io/soagen'''
                )
            )
            for annotation in self.annotations:
                o(annotation)
            with ClassDefinition(
                o,
                f'class {" ".join(self.attributes)} {self.name}',
                hidden_base_classes=[
                    rf'public soagen::mixins::size_and_capacity<{self.name}>',
                    rf'public soagen::mixins::resizable<{self.name}>',
                    rf'public soagen::mixins::equality_comparable<{self.name}>',
                    rf'public soagen::mixins::less_than_comparable<{self.name}>',
                    rf'public soagen::mixins::data_ptr<{self.name}>',
                    rf'public soagen::mixins::columns<{self.name}>',
                    rf'public soagen::mixins::rows<{self.name}>',
                    rf'public soagen::mixins::iterators<{self.name}>',
                    rf'public soagen::mixins::spans<{self.name}>',
                ]
                + ([rf'public soagen::mixins::swappable<{self.name}>'] if self.swappable else []),
            ):
                with Public(o):
                    o(
                        rf'''
                    {doxygen(r"@brief The unsigned integer size type used by this class.")}
                    using size_type = std::size_t;

                    {doxygen(r"@brief The signed integer difference type used by this class.")}
                    using difference_type = std::ptrdiff_t;

                    {doxygen(r"@brief The allocator type used by this class.")}
                    using allocator_type = soagen::allocator_type<{self.name}>;

                    {doxygen(r"@brief This class's underlying soagen::table type.")}
                    using table_type = soagen::table_type<{self.name}>;

                    {doxygen(r"@brief The soagen::table_traits for the underlying table.")}
                    using table_traits = soagen::table_traits_type<{self.name}>;

                    {doxygen(r"@brief The number of columns in the table.")}
                    static constexpr size_type column_count = table_traits::column_count;

                    {doxygen(r"@brief Gets the soagen::column_traits for a specific column of the table.")}
                    template <auto Column>
                    using column_traits = typename table_traits::template column<static_cast<size_type>(Column)>;

                    {doxygen(r"@brief Gets the type of a specific column in the table.")}
                    template <auto Column>
                    using column_type = typename column_traits<static_cast<size_type>(Column)>::value_type;

                    {doxygen(r"@brief Row iterators returned by iterator functions.")}
                    using iterator = soagen::iterator_type<{self.name}>;

                    {doxygen(r"@brief Row iterators returned by rvalue-qualified iterator functions.")}
                    using rvalue_iterator = soagen::iterator_type<{self.name}&&>;

                    {doxygen(r"@brief Row iterators returned by const-qualified iterator functions.")}
                    using const_iterator = soagen::const_iterator_type<{self.name}>;

                    {doxygen(r"@brief Regular (lvalue-qualified) span type.")}
                    using span_type = soagen::span_type<{self.name}>;

                    {doxygen(r"@brief Rvalue-qualified span type.")}
                    using rvalue_span_type = soagen::span_type<{self.name}&&>;

                    {doxygen(r"@brief Const-qualified span type.")}
                    using const_span_type = soagen::const_span_type<{self.name}>;
                    '''
                    )

                    if self.reverse_iterators:
                        o(
                            rf'''
                        {doxygen(r"@brief Reverse lvalue row iterator.")}
                        using reverse_iterator = std::reverse_iterator<iterator>;

                        {doxygen(r"@brief Reverse const lvalue row iterator.")}
                        using const_reverse_iterator = std::reverse_iterator<const_iterator>;

                        {doxygen(r"@brief Reverse rvalue row iterator.")}
                        using rvalue_reverse_iterator = std::reverse_iterator<rvalue_iterator>;
                        '''
                        )

                    o(
                        rf'''
                    {doxygen(r"@brief Regular (lvalue-qualified) row type used by this class.")}
                    using row_type = soagen::row_type<{self.name}>;

                    {doxygen(r"@brief Const row type used by this class.")}
                    using const_row_type = soagen::row_type<const {self.name}>;

                    {doxygen(r"@brief Rvalue row type used by this class.")}
                    using rvalue_row_type = soagen::row_type<{self.name}&&>;

                    {doxygen(r"""
                    @brief   The number of rows to advance to maintain the requested `alignment` for every column.

                    @details The stride size you need to use when iterating through rows of this table such that
                             the starting element for each batch in each column would have the same memory alignment as the
                             value specified for the column-specific `alignment`.

                    @note    Typically you can ignore this; column elements are always aligned correctly according to their
                             type. This is for over-alignment scenarios where you need to do things in batches (e.g. SIMD).""")}
                    static constexpr size_type aligned_stride = table_traits::aligned_stride;

                    '''
                    )

                    if o.doxygen:
                        o(
                            rf'''
                         #if SOAGEN_DOXYGEN
                        /// @brief Named index constants for all of the columns in the table.
                        using columns = POXY_IMPLEMENTATION_DETAIL(struct dummy_t);
                        #else'''
                        )

                    with ClassDefinition(o, 'enum class columns : size_type'):
                        for col in self.columns:
                            o(rf'{col.name} = {col.index},')

                    if o.doxygen:
                        o('#endif')

                    o(
                        rf'''

                    {doxygen(r"@brief Gets the name of the specified column as a null-terminated string.")}
                    template <auto Column> static constexpr auto& column_name = soagen::detail::column_name<{self.name}, static_cast<size_type>(Column)>::value;

                    {self.header}

                    '''
                    )

                for access in (r'public', r'protected', r'private'):
                    if self.static_variables[access]:
                        o()
                        o.cpp_access_level = access
                        for var in self.static_variables[access]:
                            var.write(o)

                with Private(o):
                    o(
                        rf'''
                    table_type table_;
                    '''
                    )

                with Public(o):
                    ctor_attrs = 'SOAGEN_NODISCARD_CTOR'

                    if isinstance(self.default_constructible, bool) or self.default_constructible != 'auto':
                        o(
                            rf'''
                        {doxygen(r"@brief Default constructor.")}
                        {ctor_attrs if self.default_constructible else ""} {self.name}() = {"default" if self.default_constructible else "delete"};
                        '''
                        )

                    o(
                        rf'''
                    {doxygen(r"@brief Move constructor.")}
                    {ctor_attrs if self.movable else ""} {self.name}({self.name}&&) = {"default" if self.movable else "delete"};

                    {doxygen(r"@brief Move-assignment operator.")}
                    {self.name}& operator=({self.name}&&) = {"default" if self.movable else "delete"};

                    {doxygen(r"@brief Copy constructor.")}
                    {ctor_attrs if self.copyable else ""} {self.name}(const {self.name}&) = {"default" if self.copyable else "delete"};

                    {doxygen(r"@brief Copy-assignment operator.")}
                    {self.name}& operator=(const {self.name}&) = {"default" if self.copyable else "delete"};

                    {doxygen(r"@brief Destructor.")}
                    ~{self.name}() = default;

                    {doxygen(r"@brief Constructs with the given allocator.")}
                    {ctor_attrs}
                    constexpr explicit {self.name}(const allocator_type& alloc) noexcept //
                        : table_{{ alloc }}
                    {{}}

                    {doxygen(r"@brief Constructs with the given allocator.")}
                    {ctor_attrs}
                    constexpr explicit {self.name}(allocator_type&& alloc) noexcept //
                        : table_{{ static_cast<allocator_type&&>(alloc) }}
                    {{}}

                    {doxygen(r"@brief Returns the allocator being used by the table.")}
                    SOAGEN_INLINE_GETTER
                    SOAGEN_CPP20_CONSTEXPR
                    allocator_type get_allocator() const noexcept
                    {{
                        return table_.get_allocator();
                    }}

                    '''
                    )

                # note: most of the doxygen member groups here are based on those from the cppreference.com page for
                # std::vector: https://en.cppreference.com/w/cpp/container/vector

                with DoxygenMemberGroup(o, 'Underlying table'):
                    with Public(o):
                        o(
                            rf'''
                        {doxygen(r"@brief Returns an lvalue reference to the underlying soagen::table.")}
                        SOAGEN_PURE_INLINE_GETTER
                        constexpr table_type& table() & noexcept
                        {{
                            return table_;
                        }}

                        {doxygen(r"@brief Returns an rvalue reference to the underlying soagen::table.")}
                        SOAGEN_PURE_INLINE_GETTER
                        constexpr table_type&& table() && noexcept
                        {{
                            return static_cast<table_type&&>(table_);
                        }}

                        {doxygen(r"@brief Returns a const lvalue reference to the underlying soagen::table.")}
                        SOAGEN_PURE_INLINE_GETTER
                        constexpr const table_type& table() const & noexcept
                        {{
                            return table_;
                        }}

                        {doxygen(r"@brief Returns an lvalue reference to the underlying soagen::table.")}
                        SOAGEN_PURE_INLINE_GETTER
                        explicit constexpr operator table_type& () noexcept
                        {{
                            return table_;
                        }}

                        {doxygen(r"@brief Returns an rvalue reference to the underlying soagen::table.")}
                        SOAGEN_PURE_INLINE_GETTER
                        explicit constexpr operator table_type&& () noexcept
                        {{
                            return static_cast<table_type&&>(table_);
                        }}

                        {doxygen(r"@brief Returns a const lvalue reference to the underlying soagen::table.")}
                        SOAGEN_PURE_INLINE_GETTER
                        explicit constexpr operator const table_type& () const noexcept
                        {{
                            return table_;
                        }}
                        '''
                        )

                with DoxygenMemberGroup(o, 'Modifiers'):
                    with Public(o):
                        o(
                            rf'''
                        {doxygen(r"""
                        @brief Erases the row at the given position.

                        @availability This method is only available when all the column types are move-assignable.""")}
                        SOAGEN_HIDDEN(template <bool sfinae = soagen::has_erase_member<table_type>>)
                        SOAGEN_ALWAYS_INLINE
                        SOAGEN_CPP20_CONSTEXPR
                        SOAGEN_ENABLE_IF_T({self.name}&, sfinae) erase(size_type pos) //
                            noexcept(soagen::has_nothrow_erase_member<table_type>)
                        {{
                            table_.erase(pos);
                            return *this;
                        }}

                        {doxygen(r"""
                        @brief	Erases the row at the given position without preserving order.

                        @details	This is much faster than #erase() because it uses the swap-and-pop idiom:
                                    Instead of shifting all the higher rows downward, the last row is moved into the
                                    position of the erased one and the size of the table is reduced by 1.

                        @note		If you are tracking row indices in some other place and need to maintain that invariant,
                                    you can use the return value to update your data accordingly.

                        @returns	The position of the row that was moved into the erased row's position, if any.

                        @availability This method is only available when all the column types are move-assignable.""")}
                        SOAGEN_HIDDEN(template <bool sfinae = soagen::has_unordered_erase_member<table_type>>)
                        SOAGEN_ALWAYS_INLINE
                        SOAGEN_CPP20_CONSTEXPR
                        SOAGEN_ENABLE_IF_T(soagen::optional<size_type>, sfinae) unordered_erase(size_type pos) //
                            noexcept(soagen::has_nothrow_unordered_erase_member<table_type>)
                        {{
                            return table_.unordered_erase(pos);
                        }}
                        '''
                        )

                        for const in ('', 'const_'):
                            o(
                                rf'''
                                {doxygen(rf"""
                                @brief Erases the row at the given iterator.

                                @returns    An iterator to the row immediately following the one which was removed,
                                            or #{"c" if const else ""}end() if the one removed was the last row in the table.

                                @availability This method is only available when all the column types are move-assignable.""")}
                                SOAGEN_HIDDEN(template <bool sfinae = soagen::has_erase_member<table_type>>)
                                SOAGEN_ALWAYS_INLINE
                                SOAGEN_CPP20_CONSTEXPR
                                SOAGEN_ENABLE_IF_T({const}iterator, sfinae) erase({const}iterator pos) //
                                    noexcept(soagen::has_nothrow_erase_member<table_type>)
                                {{
                                    table_.erase(static_cast<size_type>(pos));
                                    return pos;
                                }}

                                {doxygen(r"""
                                @brief	Erases the row at the given position without preserving order.

                                @details	This is much faster than #erase() because it uses the swap-and-pop idiom:
                                            Instead of shifting all the higher rows downward, the last row is moved into the
                                            position of the erased one and the size of the table is reduced by 1.

                                @note		If you are tracking row indices in some other place and need to maintain that invariant,
                                            you can use the return value to update your data accordingly.

                                @returns	The position of the row that was moved into the erased row's position, if any.

                                @availability This method is only available when all the column types are move-assignable.""")}
                                SOAGEN_HIDDEN(template <bool sfinae = soagen::has_unordered_erase_member<table_type>>)
                                SOAGEN_ALWAYS_INLINE
                                SOAGEN_CPP20_CONSTEXPR
                                SOAGEN_ENABLE_IF_T(soagen::optional<{const}iterator>, sfinae) unordered_erase({const}iterator pos) //
                                    noexcept(soagen::has_nothrow_unordered_erase_member<table_type>)
                                {{
                                    if (auto moved_pos = table_.unordered_erase(static_cast<size_type>(pos)); moved_pos)
                                        return {const}iterator{{ *this, static_cast<difference_type>(*moved_pos) }};
                                    return {{}};
                                }}
                            '''
                            )

                        o(
                            rf'''
                        {doxygen(r"""
                        @brief Swaps two columns.

                        @availability The two columns must have the same underlying value_type.""")}
                        template <auto A, auto B>
                        SOAGEN_ALWAYS_INLINE
                        SOAGEN_CPP20_CONSTEXPR
                        {self.name}& swap_columns() //
                            noexcept(noexcept(std::declval<table_type&>().template swap_columns<static_cast<size_type>(A), static_cast<size_type>(B)>()))
                        {{
                            table_.template swap_columns<static_cast<size_type>(A), static_cast<size_type>(B)>();
                            return *this;
                        }}

                        #if SOAGEN_DOXYGEN

                        {doxygen(r"@brief Removes the last row(s) from the table.")}
                        {self.name}& pop_back(size_type num = 1) noexcept(...);

                        {doxygen("@brief Removes all rows from table.")}
                        {self.name}& clear() noexcept;

                        {doxygen(r"""
                        @brief Resizes the table to the given number of rows.

                        @availability This method is only available when all the column types are default-constructible.""")}
                        {self.name}& resize(size_type new_size) //
                            noexcept(soagen::has_nothrow_resize_member<table_type>);

                        '''
                        )

                        if self.swappable:
                            o(
                                rf'''
                            {doxygen(r"""
                            @brief Swaps the contents of the table with another.

                            @availability This method is only available when #allocator_type is swappable or non-propagating.""")}
                            constexpr void swap({self.name}& other) //
                                noexcept(soagen::has_nothrow_swap_member<table_type>);

                            #endif
                            '''
                            )

                with Public(o):
                    for funcs, group, availability in (
                        ((r'push_back', r'emplace_back'), r'Adding rows', ''),
                        (
                            (r'insert', r'emplace'),
                            r'Inserting rows',
                            r'These overloads are only available when all the column types are move-constructible and move-assignable.',
                        ),
                    ):
                        sfinae_variant = -1
                        sfinae_sig_hash = ''
                        with DoxygenMemberGroup(o, group, availability=availability):
                            for func in funcs:
                                for tuple_overload in (False, True):
                                    if tuple_overload and func not in (r'emplace_back', r'emplace'):
                                        continue
                                    for rvalue_overload in (False, True):
                                        if rvalue_overload and (
                                            tuple_overload or func in (r'emplace_back', r'emplace')
                                        ):
                                            continue
                                        for iterator_overload in (None, r'iterator', r'const_iterator'):
                                            if iterator_overload and func in (r'push_back', r'emplace_back'):
                                                continue

                                            template_params = []
                                            template_defaults = []
                                            types = []
                                            names = []
                                            defaults = []
                                            forwards = []
                                            deduced = []
                                            sig_hash = func

                                            if func in (r'insert', r'emplace'):
                                                defaults.append('')
                                                if iterator_overload:
                                                    types.append(iterator_overload)
                                                    names.append(r'iter_')
                                                    forwards.append(r'static_cast<size_type>(iter_)')
                                                    deduced.append(False)
                                                else:
                                                    types.append(r'size_type')
                                                    names.append(r'index_')
                                                    forwards.append(names[-1])
                                                    deduced.append(False)

                                            if tuple_overload:
                                                template_params.append(r'typename Tuple')
                                                template_defaults += ['']
                                                types.append(r'Tuple&&')
                                                names.append(r'tuple_')
                                                deduced.append(True)
                                                defaults.append('')
                                                forwards.append(rf'static_cast<{types[-1]}>({names[-1]})')
                                            else:
                                                for col in self.columns:
                                                    names.append(col.name)
                                                    defaults.append(col.default)
                                                    if func in (r'emplace_back', r'emplace'):
                                                        typename = utils.to_pascal_case(col.name)
                                                        if typename == col.name:
                                                            typename = rf'{typename}T'
                                                        template_params.append(rf'typename {typename}')
                                                        template_defaults.append(
                                                            rf'column_traits<{col.index}>::default_emplace_type'
                                                            if col.default
                                                            else ''
                                                        )
                                                        types.append(rf'{typename}&&')
                                                        forwards.append(rf'static_cast<{types[-1]}>({names[-1]})')
                                                        deduced.append(True)
                                                    else:
                                                        category = r'rvalue' if rvalue_overload else r'param'
                                                        types.append(rf'column_traits<{col.index}>::{category}_type')
                                                        forward_type = (
                                                            rf'column_traits<{col.index}>::{category}_forward_type'
                                                        )
                                                        forwards.append(rf'static_cast<{forward_type}>({names[-1]})')
                                                        deduced.append(False)

                                            sig_hash += '##'.join(types)
                                            if not sfinae_sig_hash or sfinae_sig_hash != sig_hash:
                                                sfinae_variant = 0 if rvalue_overload else -1
                                                sfinae_sig_hash = sig_hash

                                            sfinae = []
                                            sfinae_emitted = False
                                            sfinae_template_dependent = False
                                            sfinae_condition_string = ''
                                            if tuple_overload:
                                                sfinae.append(r'table_traits::row_constructible_from<Tuple&&>')
                                                sfinae_template_dependent = True
                                            if rvalue_overload:
                                                sfinae.append(r'table_traits::rvalue_type_list_is_distinct')
                                            if func in (r'insert', r'emplace'):
                                                sfinae.append(r'table_traits::all_move_constructible')
                                                sfinae.append(r'table_traits::all_move_assignable')
                                            if func in (r'emplace_back', r'emplace'):
                                                sfinae.append(
                                                    rf'table_traits::row_constructible_from<{",".join(types[-len(self.columns):])}>'
                                                )
                                                sfinae_template_dependent = True
                                            hidden_template = bool(sfinae) and not template_params
                                            if sfinae:
                                                if len(sfinae) > 2:
                                                    sfinae_condition_string = f'\n{o.indent_str*7}&& '.join(sfinae)
                                                else:
                                                    sfinae_condition_string = r' && '.join(sfinae)
                                                if len(sfinae) >= 2:
                                                    sfinae_condition_string = rf'({sfinae_condition_string})'
                                                if sfinae_template_dependent:
                                                    template_params.append(
                                                        rf'SOAGEN_ENABLE_IF({sfinae_condition_string}'
                                                    )
                                                    template_defaults.append(')')
                                                    sfinae_emitted = True
                                                else:
                                                    if hidden_template:
                                                        template_params.append(r'bool sfinae')
                                                        template_defaults.append(sfinae_condition_string)
                                                    else:
                                                        template_params.append(
                                                            rf'SOAGEN_HIDDEN_PARAM(bool sfinae = {sfinae_condition_string}'
                                                        )
                                                        template_defaults.append(')')

                                            defaults_ok = True
                                            assert len(template_params) == len(template_defaults)
                                            for i in range(len(template_defaults) - 1, -1, -1):
                                                if not defaults_ok:
                                                    template_defaults[i] = ''
                                                elif not template_defaults[i]:
                                                    defaults_ok = False

                                            defaults_ok = True
                                            assert len(names) == len(defaults)
                                            assert len(names) == len(types)
                                            for i in range(len(defaults) - 1, -1, -1):
                                                if not defaults_ok:
                                                    defaults[i] = ''
                                                elif not defaults[i]:
                                                    defaults_ok = False

                                            s = ''

                                            # doxygen
                                            if o.doxygen:
                                                s += '/// @brief '
                                                if func in (r'emplace_back', r'emplace'):
                                                    s += 'Constructs a new row directly in-place'
                                                elif func == r'insert':
                                                    s += 'Inserts a new row'
                                                elif func == r'push_back':
                                                    s += 'Adds a new row'
                                                if func in (r'insert', r'emplace'):
                                                    s += ' at an arbitrary position in the table'
                                                elif func.endswith(r'_back'):
                                                    s += ' at the end of the table'
                                                if tuple_overload:
                                                    s += ' by unpacking a tuple-like object'
                                                if rvalue_overload:
                                                    s += ' (rvalue overload)'
                                                s += '.\n'
                                                if func in (r'insert', r'emplace'):
                                                    s += '///\n'
                                                    s += '/// @availability This overload is only available when all'
                                                    s += ' the column types are move-constructible and move-assignable.\n'

                                            # template <>
                                            if template_params:
                                                if hidden_template:
                                                    s += r'SOAGEN_HIDDEN('
                                                s += r'template <'
                                                padding = f'\n{TAB_SPACES*2}  ' if len(template_params) >= 3 else ' '
                                                for i in range(len(template_params)):
                                                    if i:
                                                        if not template_params[i].startswith(
                                                            'SOAGEN_ENABLE_IF'
                                                        ) and not template_params[i].startswith('SOAGEN_HIDDEN'):
                                                            s += rf','
                                                        s += rf'{padding}'
                                                    s += template_params[i]
                                                    if template_defaults[i] == ')':
                                                        s += ')'
                                                    elif template_defaults[i]:
                                                        s += rf' = {template_defaults[i]}'
                                                s += r'>'
                                                if hidden_template:
                                                    s += r')'
                                                s += '\n'

                                            # attributes
                                            s += 'SOAGEN_CPP20_CONSTEXPR\n'

                                            # return type
                                            return_type = iterator_overload if iterator_overload else rf'{self.name}&'
                                            return_expression = r'iter_' if iterator_overload else r'*this'
                                            if sfinae and not sfinae_emitted and sfinae_variant == -1:
                                                return_type = rf'SOAGEN_ENABLE_IF_T({return_type}, sfinae)'
                                                sfinae_emitted = True
                                                sfinae_variant += 1
                                            s += rf'{return_type} '

                                            # name + args
                                            s += rf'{func}('
                                            if types:
                                                padding = ' ' * (len(func) + len(return_type) + 2)
                                                padding = f'\n{padding}' if len(types) >= 3 else ' '
                                                for i in range(len(types)):
                                                    if i:
                                                        s += rf',{padding}'
                                                    if (
                                                        sfinae
                                                        and not sfinae_emitted
                                                        and sfinae_variant == i
                                                        and types[i]
                                                        and not deduced[i]
                                                        in (
                                                            'size_t',
                                                            'size_type',
                                                            'iterator',
                                                            'const_iterator',
                                                            rf'column_traits<{i-1}>::param_type',
                                                            rf'column_traits<{i-1}>::rvalue_type',
                                                        )
                                                    ):
                                                        s += rf'SOAGEN_ENABLE_IF_T({types[i]}, sfinae)'
                                                        sfinae_emitted = True
                                                        sfinae_variant += 1
                                                    else:
                                                        s += rf'{types[i]}'
                                                    s += rf' {names[i]}'
                                                    if defaults[i]:
                                                        s += rf' = {defaults[i]}'
                                            s += ') //\n'

                                            # noexcept(...)
                                            backing_func = (
                                                r'emplace' if func in (r'insert', r'emplace') else r'emplace_back'
                                            )
                                            s += '\tnoexcept('
                                            if func == 'push_back':
                                                if tuple_overload:
                                                    s += rf'table_traits::row_push_back_is_nothrow<table_type&, {types[0]}>'
                                                else:
                                                    s += rf'table_traits::{"rvalue_" if rvalue_overload else ""}push_back_is_nothrow<table_type&>'
                                            elif func == 'insert':
                                                if tuple_overload:
                                                    s += rf'table_traits::row_insert_is_nothrow<table_type&, {types[-1]}>'
                                                else:
                                                    s += rf'table_traits::{"rvalue_" if rvalue_overload else ""}insert_is_nothrow<table_type&>'
                                            elif func == 'emplace_back':
                                                s += rf'table_traits::emplace_back_is_nothrow<table_type&, {", ".join(types)}>'
                                            elif func == 'emplace':
                                                s += rf'table_traits::emplace_is_nothrow<table_type&, {", ".join(types[1:])}>'
                                            else:
                                                assert False  # we should have all our bases covered by the table_traits
                                            s += ')\n'

                                            # requires(...)
                                            if sfinae:
                                                s += f'\tSOAGEN_REQUIRES({sfinae_condition_string.strip("()")}) //\n'

                                            # function body
                                            s += '{\n'
                                            s += f'\ttable_.{backing_func}('
                                            if forwards:
                                                padding = f'\n{TAB_SPACES*2}' if len(forwards) >= 3 else ' '
                                                if len(forwards) >= 3:
                                                    s += padding
                                                s += f',{padding}'.join(forwards)
                                            s += f');\n'
                                            if return_expression:
                                                s += f'\treturn {return_expression};\n'
                                            s += '}\n\n'

                                            assert not sfinae or sfinae_emitted
                                            o(s)

                with DoxygenMemberGroup(o, 'Columns'):
                    with Public(o):
                        o(
                            rf'''
                        {doxygen(r"@brief Returns a pointer to the elements of a specific column.")}
                        template <auto Column>
                        SOAGEN_COLUMN({self.name}, Column)
                        constexpr column_type<Column>* column() noexcept
                        {{
                            static_assert(static_cast<size_type>(Column) < table_traits::column_count, "column index out of range");

                            return soagen::assume_aligned<soagen::actual_alignment<{self.name}, Column>>(table_.template column<Column>());
                        }}

                        {doxygen(r"@brief Returns a pointer to the elements of a specific column.")}
                        template <auto Column>
                        SOAGEN_COLUMN({self.name}, Column)
                        constexpr std::add_const_t<column_type<Column>>* column() const noexcept
                        {{
                            static_assert(static_cast<size_type>(Column) < table_traits::column_count, "column index out of range");

                            return soagen::assume_aligned<soagen::actual_alignment<{self.name}, Column>>(table_.template column<Column>());
                        }}

                        '''
                        )

                        if o.doxygen:
                            o(
                                rf'''
                            #if SOAGEN_DOXYGEN

                            {doxygen(r"""
                            @brief Returns a pointer to the raw byte backing array.

                            @availability This method is only available when all the column types are trivially-copyable.""")}
                            constexpr std::byte* data() noexcept;

                            {doxygen(r"""
                            @brief Returns a pointer to the raw byte backing array.

                            @availability This method is only available when all the column types are trivially-copyable.""")}
                            constexpr const std::byte* const data() noexcept;

                            '''
                            )
                            for i in range(len(self.columns)):
                                o(
                                    rf'''
                                {doxygen(rf"@brief Returns a pointer to the elements in column [{i}]: {self.columns[i].name}.")}
                                constexpr {self.columns[i].pointer_type} {self.columns[i].name}() noexcept;

                                {doxygen(rf"@brief Returns a pointer to the elements in column [{i}]: {self.columns[i].name}.")}
                                constexpr {self.columns[i].const_pointer_type} {self.columns[i].name}() const noexcept;
                                '''
                                )
                            for const in ('', 'const '):
                                o(
                                    rf'''
                                {doxygen(r"""
                                @brief Invokes a function once for each column data pointer.

                                @tparam Func A callable type compatible with one of the following signatures:<ul>
                                        <li> `void(auto*, std::integral_constant<size_type, N>)`
                                        <li> `void(auto*, size_type)`
                                        <li> `void(std::integral_constant<size_type, N>, auto*)`
                                        <li> `void(size_type, auto*)`
                                        <li> `void(auto*)`
                                </ul>
                                Overload resolution is performed in the order listed above.

                                @param func The callable to invoke.""")}
                                template <typename Func>
                                constexpr void for_each_column(Func&& func) {const} noexcept(...);
                                '''
                                )
                            o(
                                '''

                            #endif

                            '''
                            )

                if o.doxygen:  # everything here is exposition only
                    with PreprocessorRegion(o, 'SOAGEN_DOXYGEN'), Public(o):
                        # Capacity
                        with DoxygenMemberGroup(o, 'Capacity'):
                            with Public(o):
                                o(
                                    rf'''
                                {doxygen(r"@brief Returns true if the number of rows is zero.")}
                                SOAGEN_PURE_INLINE_GETTER
                                constexpr bool empty() const noexcept;

                                {doxygen(r"@brief Returns the current number of rows.")}
                                SOAGEN_PURE_INLINE_GETTER
                                constexpr size_type size() const noexcept;

                                {doxygen(r"@brief Returns the maximum possible number of rows.")}
                                SOAGEN_PURE_INLINE_GETTER
                                constexpr size_type max_size() const noexcept;

                                {doxygen(r"""
                                @brief Returns the size of the current underlying buffer allocation in bytes.

                                @warning This value is `capacity() * (sizeof() for every column) + (alignment padding)`.
                                        It is **not** based on `size()`! If you are using the value returned by this function
                                        in conjunction with `data()` to do serialization, hashing, etc, use `shrink_to_fit()` first.""")}
                                constexpr size_type allocation_size() const noexcept;

                                {doxygen(r"@brief Reserves storage for (at least) the given number of rows.")}
                                {self.name}& reserve(size_type new_cap)noexcept(...)

                                {doxygen(r"@brief Returns the number of rows that can be held in currently allocated storage.")}
                                constexpr size_type capacity() const noexcept;

                                {doxygen(r"@brief Frees unused capacity.")}
                                {self.name}& shrink_to_fit() noexcept(...);
                                '''
                                )

                        # Equality
                        with DoxygenMemberGroup(
                            o,
                            'Equality',
                            availability=r'These operators are only available when all the column types are equality-comparable.',
                        ):
                            o(
                                rf'''
                            {doxygen(r"""
                            @brief Returns true if all of the elements in two tables are equal.""")}
                            friend constexpr bool operator==(const {self.name}& lhs, const {self.name}& rhs) //
                                noexcept(soagen::is_nothrow_equality_comparable<table_type>);

                            {doxygen(r"""
                            @brief Returns true if not all of the elements in two tables are equal.""")}
                            friend constexpr bool operator!=(const {self.name}& lhs, const {self.name}& rhs) //
                                noexcept(soagen::is_nothrow_equality_comparable<table_type>);
                            '''
                            )

                        # Comparsion
                        with DoxygenMemberGroup(
                            o,
                            'Comparison',
                            availability=r'These operators are only available when all the column types are less-than-comparable.',
                        ):
                            o(
                                rf'''
                            {doxygen("""
                            @brief Returns true if the LHS table is ordered lexicographically less-than the RHS table.""")}
                            friend constexpr bool operator<(const {self.name}& lhs, const {self.name}& rhs) //
                                noexcept(soagen::is_nothrow_less_than_comparable<table_type>);

                            {doxygen("""
                            @brief Returns true if the LHS table is ordered lexicographically less-than-or-equal-to the RHS table.""")}
                            friend constexpr bool operator<=(const {self.name}& lhs, const {self.name}& rhs) //
                                noexcept(soagen::is_nothrow_less_than_comparable<table_type>);

                            {doxygen("""
                            @brief Returns true if the LHS table is ordered lexicographically greater-than the RHS table.""")}
                            friend constexpr bool operator>(const {self.name}& lhs, const {self.name}& rhs) //
                                noexcept(soagen::is_nothrow_less_than_comparable<table_type>);

                            {doxygen("""
                            @brief Returns true if the LHS table is ordered lexicographically greater-than-or-equal-to the RHS table.""")}
                            friend constexpr bool operator>=(const {self.name}& lhs, const {self.name}& rhs) //
                                noexcept(soagen::is_nothrow_less_than_comparable<table_type>);
                            '''
                            )

                        # Iterators
                        for funcs, group in ((('', 'c'), r'Iterators'), (('r', 'cr'), r'Iterators (reverse)')):
                            reverse = 'r' in funcs
                            if reverse and not self.reverse_iterators:
                                continue
                            with DoxygenMemberGroup(o, group):
                                for func in funcs:
                                    for const, ref in (('', '&'), ('', '&&'), ('const ', '&')):
                                        if func.startswith('c') and (not const or ref == '&&'):
                                            continue

                                        func_ref = ref if not func.startswith('c') else ''
                                        an = 'a reverse' if reverse else 'an'
                                        past = 'before' if reverse else 'past'
                                        first = 'last' if reverse else 'first'
                                        last = 'first' if reverse else 'last'
                                        iterator = rf'soagen::{"const_" if const else ""}iterator_type<{self.name}{"&&" if ref == "&&" else ""}, Columns...>'

                                        o(
                                            rf'''

                                        {doxygen(rf"""@brief Returns {an} iterator to the {first} row in the table.""")}
                                        template <auto... Columns>
                                        constexpr {iterator} {func}begin() {const}{func_ref} noexcept;

                                        {doxygen(rf"""@brief Returns {an} iterator to one-{past}-the-{last} row in the table.""")}
                                        template <auto... Columns>
                                        constexpr {iterator} {func}end() {const}{func_ref} noexcept;
                                        '''
                                        )

                        # Rows
                        with DoxygenMemberGroup(o, 'Rows'):
                            for const, ref in (('', '&'), ('', '&&'), ('const ', '&')):
                                row_type = ('rvalue_' if ref == '&&' else ('const_' if const else '')) + r'row_type'
                                o(
                                    rf'''
                                    {doxygen(r"""
                                    @brief Returns the row at the given index.

                                    @tparam Columns Indices of the columns to include in the row. Leave the list empty for all columns.""")}
                                    template <auto... Columns>
                                    soagen::row_type<{const}{self.name}{'&&' if ref == '&&' else ''}, Columns...> row(size_type index) {const}{ref} noexcept;

                                    {doxygen(r"""@brief Returns the row at the given index.""")}
                                    {row_type} operator[](size_type index) {const}{ref} noexcept;

                                    {doxygen(r"""
                                    @brief Returns the row at the given index.

                                    @throws std::out_of_range""")}
                                    {row_type} at(size_type index) {const}{ref};

                                    {doxygen(r"""@brief Returns the very first row in the table.""")}
                                    {row_type} front() {const}{ref} noexcept;

                                    {doxygen(r"""@brief Returns the very last row in the table.""")}
                                    {row_type} back() {const}{ref} noexcept;

                                '''
                                )

                        # Spans
                        with DoxygenMemberGroup(o, 'Spans'):
                            o(
                                rf'''
                                {doxygen(rf"""@brief Returns a span of (some part of) the table.""")}
                                span_type subspan(size_type start, size_type count = static_cast<size_type>(-1)) & noexcept;

                                {doxygen(rf"""@brief Returns an rvalue-qualified span of (some part of) the table.""")}
                                rvalue_span_type subspan(size_type start, size_type count = static_cast<size_type>(-1)) && noexcept;

                                {doxygen(rf"""@brief Returns a const-qualified span of (some part of) the table.""")}
                                const_span_type subspan(size_type start, size_type count = static_cast<size_type>(-1)) const& noexcept;

                                {doxygen(rf"""@brief Returns a const-qualified span of (some part of) the table.""")}
                                const_span_type const_subspan(size_type start, size_type count = static_cast<size_type>(-1)) const noexcept;
                                '''
                            )
                o(
                    rf'''
                {self.footer}
                '''
                )

            if self.swappable:
                o(
                    rf'''
                {doxygen(f"""
                @brief Swaps the contents of two instances of #{self.qualified_name}.

                @availability	This overload is only available when #{self.qualified_name}::allocator_type
                                is swappable or non-propagating.""")}
                SOAGEN_HIDDEN_CONSTRAINT(sfinae, auto sfinae = soagen::has_swap_member<{self.name}>)
                SOAGEN_ALWAYS_INLINE
                constexpr void swap({self.name}& lhs, {self.name}& rhs) //
                    noexcept(soagen::has_nothrow_swap_member<{self.name}>)
                {{
                    lhs.swap(rhs);
                }}
                '''
                )

            if self.epilogue:
                o(
                    f'''
                {self.epilogue}
                '''
                )

    def write_outline_member_implementations(self, o: Writer):
        pass


__all__ = [r'Struct']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

"""
Utility classes to build stacks of 'meta variables' for text substitution.
"""

from . import utils
from .configurable import Configurable


class MetaVars(object):

    """
    A collection of meta-variables localized to one context (e.g. a class scope).
    """

    def __init__(self):
        self.__vars = dict()

    def push(self, name: str, val):
        assert val is not None
        assert name is not None
        assert not isinstance(name, MetaVars)
        assert not isinstance(val, MetaVars)
        if not isinstance(name, str):
            name = str(name)
        name = name.strip()
        assert name
        if not isinstance(val, str):
            val = str(val)
        if name not in self.__vars:
            self.__vars[name] = []
        self.__vars[name].append(val)
        return self

    def pop(self, name: str):
        assert name is not None
        assert not isinstance(name, MetaVars)
        if not isinstance(name, str):
            name = str(name)
        name = name.strip()
        assert name
        self.__vars[name].pop()
        return self

    def __call__(self, text, skip_names=None) -> str:
        for name, stack in self.__vars.items():
            if not stack or (skip_names and name in skip_names):
                continue
            text = utils.replace_metavar(name, stack[-1], text)
            if skip_names is not None:
                skip_names.add(name)
        return text


class MetaStack(object):

    """
    A stack of MetaVars objects.
    """

    def __init__(self):
        self.__vars = []

    def push(self, vars: MetaVars):
        assert vars is not None
        assert isinstance(vars, MetaVars)
        self.__vars.append(vars)

    def pop(self):
        if self.__vars:
            self.__vars.pop()

    def __call__(self, text: str) -> str:
        if not self.__vars:
            return text
        if len(self.__vars) == 1:
            return self.__vars[0](text)

        skip_names = set()
        for i in range(len(self.__vars) - 1, -1, -1):
            text = self.__vars[i](text, skip_names)
        return text


class MetaScope(object):

    """
    A scope that pushes a MetaVars object onto a MetaStack upon entry, and pops it upon exit.
    """

    def __init__(self, stack, vars=None):
        if isinstance(stack, Configurable):
            vars = stack
            stack = stack.config

        assert stack is not None
        if not isinstance(stack, MetaStack):
            if hasattr(stack, 'meta_stack'):
                stack = stack.meta_stack
        assert isinstance(stack, MetaStack)

        assert vars is not None
        if not isinstance(vars, MetaVars):
            if hasattr(vars, 'meta_vars'):
                vars = vars.meta_vars
            elif hasattr(vars, 'meta'):
                vars = vars.meta
        assert isinstance(vars, MetaVars)

        self.__stack = stack
        self.__vars = vars

    def __enter__(self):
        self.__stack.push(self.__vars)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.__stack.pop()

    def __call__(self, text) -> str:
        return self.__stack(text)


__all__ = [r'MetaVars', r'MetaStack', r'MetaScope']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import re

from trieregex import TrieRegEx


def make_regex(*symbols):
    return re.compile(rf'\b(?:{TrieRegEx(*[s.strip() for s in symbols]).regex()})\b')


# yapf: disable

HEADER_MATCHERS = (
    (r'cstddef', make_regex(
    r'std::byte',
    r'std::max_align_t',
    r'std::ptrdiff_t',
    r'std::size_t',
    )),

    (r'cstdlib', make_regex(
    r'std::div_t',
    r'std::ldiv_t',
    r'std::lldiv_t',
    )),

    (r'limits', make_regex(
    r'std::float_denorm_style',
    r'std::float_round_style',
    r'std::numeric_limits',
    )),

    (r'climits', make_regex(
    r'CHAR_BIT',
    r'CHAR_MAX',
    r'CHAR_MIN',
    r'INT_MAX',
    r'INT_MIN',
    r'LLONG_MAX',
    r'LLONG_MIN',
    r'LONG_MAX',
    r'LONG_MIN',
    r'MB_LEN_MAX',
    r'SCHAR_MAX',
    r'SCHAR_MIN',
    r'SHRT_MAX',
    r'SHRT_MIN',
    r'UCHAR_MAX',
    r'UINT_MAX',
    r'ULLONG_MAX',
    r'ULONG_MAX',
    r'USHRT_MAX',
    )),

    (r'cfloat', make_regex(
    r'DBL_DECIMAL_DIG',
    r'DBL_DIG',
    r'DBL_EPSILON',
    r'DBL_HAS_SUBNORM',
    r'DBL_MANT_DIG',
    r'DBL_MAX_10_EXP',
    r'DBL_MAX_EXP',
    r'DBL_MAX',
    r'DBL_MIN_10_EXP',
    r'DBL_MIN_EXP',
    r'DBL_MIN',
    r'DBL_TRUE_MIN',
    r'DECIMAL_DIG',
    r'FLT_DECIMAL_DIG',
    r'FLT_DIG',
    r'FLT_EPSILON',
    r'FLT_EVAL_METHOD',
    r'FLT_HAS_SUBNORM',
    r'FLT_MANT_DIG',
    r'FLT_MAX_10_EXP',
    r'FLT_MAX_EXP',
    r'FLT_MAX',
    r'FLT_MIN_10_EXP',
    r'FLT_MIN_EXP',
    r'FLT_MIN',
    r'FLT_RADIX',
    r'FLT_ROUNDS',
    r'FLT_TRUE_MIN',
    r'LDBL_DECIMAL_DIG',
    r'LDBL_DIG',
    r'LDBL_EPSILON',
    r'LDBL_HAS_SUBNORM',
    r'LDBL_MANT_DIG',
    r'LDBL_MAX_10_EXP',
    r'LDBL_MAX_EXP',
    r'LDBL_MAX',
    r'LDBL_MIN_10_EXP',
    r'LDBL_MIN_EXP',
    r'LDBL_MIN',
    r'LDBL_TRUE_MIN',
    )),

    (r'cstdint', make_regex(
    r'std::int_fast16_t',
    r'std::int_fast32_t',
    r'std::int_fast64_t',
    r'std::int_fast8_t',
    r'std::int_fastN_t',
    r'std::int_least16_t',
    r'std::int_least32_t',
    r'std::int_least64_t',
    r'std::int_least8_t',
    r'std::int16_t',
    r'std::int32_t',
    r'std::int64_t',
    r'std::int8_t',
    r'std::intmax_t',
    r'std::intptr_t',
    r'std::uint_fast16_t',
    r'std::uint_fast32_t',
    r'std::uint_fast64_t',
    r'std::uint_fast8_t',
    r'std::uint_least16_t',
    r'std::uint_least32_t',
    r'std::uint_least64_t',
    r'std::uint_least8_t',
    r'std::uint16_t',
    r'std::uint32_t',
    r'std::uint64_t',
    r'std::uint8_t',
    r'std::uintmax_t',
    r'std::uintptr_t',
    r'INT_FASTN_MAX',
    r'INT_FASTN_MIN',
    r'INT_LEASTN_MAX',
    r'INT_LEASTN_MIN',
    r'INTMAX_C',
    r'INTMAX_MAX',
    r'INTMAX_MIN',
    r'INTN_C',
    r'INTN_MAX',
    r'INTN_MIN',
    r'INTPTR_MAX',
    r'INTPTR_MIN',
    r'PTRDIFF_MAX',
    r'PTRDIFF_MIN',
    r'SIG_ATOMIC_MAX',
    r'SIG_ATOMIC_MIN',
    r'SIZE_MAX',
    r'UINT_FASTN_MAX',
    r'UINT_LEASTN_MAX',
    r'UINTMAX_C',
    r'UINTMAX_MAX',
    r'UINTN_C',
    r'UINTN_MAX',
    r'UINTPTR_MAX',
    r'WCHAR_MAX',
    r'WCHAR_MIN',
    r'WINT_MAX',
    r'WINT_MIN',
    )),

    (r'stdfloat', make_regex(
    r'std::bfloat16_t',
    r'std::float128_t',
    r'std::float16_t',
    r'std::float32_t',
    r'std::float64_t',
    )),

    (r'new', make_regex(
    r'std::align_val_t',
    r'std::bad_alloc',
    r'std::bad_array_new_length',
    r'std::destroying_delete_t',
    r'std::destroying_delete',
    r'std::hardware_constructive_interference_size',
    r'std::hardware_destructive_interference_size',
    r'std::new_handler',
    r'std::nothrow_t',
    r'std::nothrow',
    )),

    (r'typeinfo', make_regex(
    r'std::bad_cast',
    r'std::bad_typeid',
    r'std::type_info',
    )),

    (r'source_location', make_regex(
    r'std::source_location',
    )),

    (r'exception', make_regex(
    r'std::bad_exception',
    r'std::exception_ptr',
    r'std::exception',
    r'std::nested_exception',
    r'std::terminate_handler',
    )),

    (r'initializer_list', make_regex(
    r'std::initializer_list',
    )),

    (r'compare', make_regex(
    r'std::common_comparison_category_t',
    r'std::common_comparison_category',
    r'std::compare_partial_order_fallback',
    r'std::compare_strong_order_fallback',
    r'std::compare_three_way_result_t',
    r'std::compare_three_way_result',
    r'std::compare_three_way',
    r'std::compare_weak_order_fallback',
    r'std::partial_order',
    r'std::partial_ordering',
    r'std::strong_order',
    r'std::strong_ordering',
    r'std::three_way_comparable_with',
    r'std::three_way_comparable',
    r'std::weak_order',
    r'std::weak_ordering',
    )),

    (r'coroutine', make_regex(
    r'std::coroutine_handle',
    r'std::coroutine_traits',
    r'std::noop_coroutine_handle',
    r'std::noop_coroutine_promise',
    r'std::noop_coroutine',
    r'std::suspend_always',
    r'std::suspend_never',
    )),

    (r'csignal', make_regex(
    r'std::sig_atomic_t',
    r'SIG_DFL',
    r'SIG_ERR',
    r'SIG_IGN',
    r'SIGABRT',
    r'SIGFPE',
    r'SIGILL',
    r'SIGINT',
    r'SIGSEGV',
    r'SIGTERM',
    )),

    (r'concepts', make_regex(
    r'std::assignable_from',
    r'std::common_reference_with',
    r'std::common_with',
    r'std::constructible_from',
    r'std::convertible_to',
    r'std::copy_constructible',
    r'std::copyable',
    r'std::default_initializable',
    r'std::derived_from',
    r'std::destructible',
    r'std::equality_comparable_with',
    r'std::equality_comparable',
    r'std::equivalence_relation',
    r'std::floating_point',
    r'std::integral',
    r'std::invocable',
    r'std::movable',
    r'std::move_constructible',
    r'std::predicate',
    r'std::regular_invocable',
    r'std::regular',
    r'std::relation',
    r'std::same_as',
    r'std::semiregular',
    r'std::signed_integral',
    r'std::strict_weak_order',
    r'std::swappable_with',
    r'std::swappable',
    r'std::totally_ordered_with',
    r'std::totally_ordered',
    r'std::unsigned_integral',
    )),

    (r'utility', make_regex(
    r'std::as_const',
    r'std::cmp_equal',
    r'std::cmp_greater_equal',
    r'std::cmp_greater',
    r'std::cmp_less_equal',
    r'std::cmp_less',
    r'std::cmp_not_equal',
    r'std::declval',
    r'std::exchange',
    r'std::forward_like',
    r'std::forward',
    r'std::in_place_index_t',
    r'std::in_place_index',
    r'std::in_place_t',
    r'std::in_place_type_t',
    r'std::in_place_type',
    r'std::in_place',
    r'std::in_range',
    r'std::integer_sequence',
    r'std::make_pair',
    r'std::move_if_noexcept',
    r'std::move',
    r'std::pair',
    r'std::piecewise_construct_t',
    r'std::piecewise_construct',
    r'std::swap',
    r'std::to_underlying',
    r'std::unreachable',
    )),

    (r'memory', make_regex(
    r'std::allocate_shared_for_overwrite',
    r'std::allocate_shared',
    r'std::allocation_result',
    r'std::allocator_arg_t',
    r'std::allocator_arg',
    r'std::allocator_traits',
    r'std::allocator',
    r'std::assume_aligned',
    r'std::bad_weak_ptr',
    r'std::construct_at',
    r'std::default_delete',
    r'std::destroy_at',
    r'std::destroy_n',
    r'std::destroy',
    r'std::enable_shared_from_this',
    r'std::inout_ptr_t',
    r'std::make_shared_for_overwrite',
    r'std::make_shared',
    r'std::make_unique_for_overwrite',
    r'std::make_unique',
    r'std::out_ptr_t',
    r'std::owner_less',
    r'std::pointer_safety',
    r'std::pointer_traits',
    r'std::raw_storage_iterator',
    r'std::shared_ptr',
    r'std::start_lifetime_as_array',
    r'std::start_lifetime_as',
    r'std::to_address',
    r'std::uninitialized_copy_n',
    r'std::uninitialized_copy',
    r'std::uninitialized_default_construct_n',
    r'std::uninitialized_default_construct',
    r'std::uninitialized_fill_n',
    r'std::uninitialized_fill',
    r'std::uninitialized_move_n',
    r'std::uninitialized_move',
    r'std::uninitialized_value_construct_n',
    r'std::uninitialized_value_construct',
    r'std::unique_ptr',
    r'std::uses_allocator_v',
    r'std::uses_allocator',
    r'std::weak_ptr',
    )),

    (r'tuple', make_regex(
    r'std::apply',
    r'std::forward_as_tuple',
    r'std::ignore',
    r'std::make_from_tuple',
    r'std::make_tuple',
    r'std::tie',
    r'std::tuple',
    )),


    (r'optional', make_regex(
    r'std::bad_optional_access',
    r'std::make_optional',
    r'std::nullopt_t',
    r'std::nullopt',
    r'std::optional',
    )),

    (r'variant', make_regex(
    r'std::bad_variant_access',
    r'std::get_if',
    r'std::holds_alternative',
    r'std::monostate',
    r'std::variant_alternative_t',
    r'std::variant_alternative',
    r'std::variant_npos',
    r'std::variant_size_v',
    r'std::variant_size',
    r'std::variant',
    r'std::visit',
    )),

    (r'array', make_regex(
    r'std::array',
    r'std::to_array',
    )),

    (r'deque', make_regex(
    r'std::deque',
    )),

    (r'forward_list', make_regex(
    r'std::forward_list',
    )),

    (r'list', make_regex(
    r'std::list',
    )),

    (r'vector', make_regex(
    r'std::vector',
    )),

    (r'map', make_regex(
    r'std::map',
    r'std::multimap',
    )),

    (r'unordered_map', make_regex(
    r'std::unordered_map',
    r'std::unordered_multimap',
    )),

    (r'flat_map', make_regex(
    r'std::flat_map',
    r'std::flat_multimap',
    )),

    (r'set', make_regex(
    r'std::set',
    r'std::multiset',
    )),

    (r'unordered_set', make_regex(
    r'std::unordered_set',
    r'std::unordered_multiset',
    )),

    (r'flat_set', make_regex(
    r'std::flat_set',
    r'std::flat_multiset',
    )),

    (r'queue', make_regex(
    r'std::queue',
    r'std::priority_queue',
    )),

    (r'stack', make_regex(
    r'std::stack',
    )),

    (r'span', make_regex(
    r'std::as_bytes',
    r'std::as_writable_bytes',
    r'std::dynamic_extent',
    r'std::span',
    )),

    (r'mdspan', make_regex(
    r'std::default_accessor',
    r'std::extents',
    r'std::layout_left',
    r'std::layout_right',
    r'std::layout_stride',
    r'std::mdspan',
    )),

    (r'string', make_regex(
    r'std::basic_string',
    r'std::literals::string_literals',
    r'std::string_literals',
    r'std::string',
    r'std::to_string',
    r'std::to_wstring',
    r'std::u16string',
    r'std::u32string',
    r'std::u8string',
    r'std::wstring',
    r'"s'
    )),

    (r'string_view', make_regex(
    r'std::basic_string_view',
    r'std::literals::string_view_literals',
    r'std::string_view_literals',
    r'std::string_view',
    r'std::u16string_view',
    r'std::u32string_view',
    r'std::u8string_view',
    r'std::wstring_view',
    r'"sv'
    )),

    (r'iterator', make_regex(
    r'std::back_insert_iterator',
    r'std::bidirectional_iterator_tag',
    r'std::common_iterator',
    r'std::contiguous_iterator_tag',
    r'std::counted_iterator',
    r'std::forward_iterator_tag',
    r'std::front_insert_iterator',
    r'std::input_iterator_tag',
    r'std::insert_iterator',
    r'std::iterator',
    r'std::move_iterator',
    r'std::move_sentinel',
    r'std::output_iterator_tag',
    r'std::random_access_iterator_tag',
    r'std::reverse_iterator',
    )),

    (r'chrono', make_regex(
    r'std::chrono::ambiguous_local_time',
    r'std::chrono::choose',
    r'std::chrono::clock_time_conversion',
    r'std::chrono::day',
    r'std::chrono::days',
    r'std::chrono::duration_values',
    r'std::chrono::duration',
    r'std::chrono::file_clock',
    r'std::chrono::gps_clock',
    r'std::chrono::hh_mm_ss',
    r'std::chrono::high_resolution_clock',
    r'std::chrono::hours',
    r'std::chrono::is_clock_v',
    r'std::chrono::is_clock',
    r'std::chrono::last_spec',
    r'std::chrono::leap_second_info',
    r'std::chrono::leap_second',
    r'std::chrono::local_info',
    r'std::chrono::local_t',
    r'std::chrono::microseconds',
    r'std::chrono::milliseconds',
    r'std::chrono::minutes',
    r'std::chrono::month_day_last',
    r'std::chrono::month_day',
    r'std::chrono::month_weekday_last',
    r'std::chrono::month_weekday',
    r'std::chrono::month',
    r'std::chrono::months',
    r'std::chrono::nanoseconds',
    r'std::chrono::nonexistent_local_time',
    r'std::chrono::seconds',
    r'std::chrono::steady_clock',
    r'std::chrono::sys_info',
    r'std::chrono::system_clock',
    r'std::chrono::tai_clock',
    r'std::chrono::time_point',
    r'std::chrono::time_zone_link',
    r'std::chrono::time_zone',
    r'std::chrono::treat_as_floating_point',
    r'std::chrono::tzdb_list',
    r'std::chrono::tzdb',
    r'std::chrono::utc_clock',
    r'std::chrono::weekday_indexed',
    r'std::chrono::weekday_last',
    r'std::chrono::weekday',
    r'std::chrono::weeks',
    r'std::chrono::year_month_day_last',
    r'std::chrono::year_month_day',
    r'std::chrono::year_month_weekday_last',
    r'std::chrono::year_month_weekday',
    r'std::chrono::year_month',
    r'std::chrono::year',
    r'std::chrono::years',
    r'std::chrono::zoned_time',
    r'std::chrono::zoned_traits',
    )),

    (r'mz/type_list.hpp', make_regex(
    r'mz::type_list',
    )),

    (r'mz/spin_mutex.hpp', make_regex(
    r'mz::spin_mutex',
    )),

    (r'mz/tagged_ptr.hpp', make_regex(
    r'mz::tagged_ptr',
    )),

    (r'mz/noexcept.hpp', make_regex(
    r'mz::add_noexcept',
    r'mz::add_noexcept_t',
    r'mz::remove_noexcept',
    r'mz::remove_noexcept_t',
    r'mz::is_noexcept',
    r'mz::is_noexcept_v',
    )),

    (r'toml++/toml.h', make_regex(
    r'toml::array_iterator',
    r'toml::array',
    r'toml::const_array_iterator',
    r'toml::const_table_iterator',
    r'toml::date_time',
    r'toml::date',
    r'toml::default_formatter',
    r'toml::format_flags',
    r'toml::inserted_type_of',
    r'toml::inserter',
    r'toml::json_formatter',
    r'toml::key',
    r'toml::node_type',
    r'toml::node_view',
    r'toml::node',
    r'toml::optional',
    r'toml::parse_error',
    r'toml::parse_result',
    r'toml::path_component_type',
    r'toml::path_component',
    r'toml::path',
    r'toml::source_index',
    r'toml::source_path_ptr',
    r'toml::source_position',
    r'toml::source_region',
    r'toml::table_iterator',
    r'toml::table',
    r'toml::time_offset',
    r'toml::time',
    r'toml::time',
    r'toml::toml_formatter',
    r'toml::value_flags',
    r'toml::value',
    r'toml::yaml_formatter',
    )),

    (r'muu/span.h', make_regex(
    r'muu::span',
    r'muu::const_span',
    )),

    (r'tcb/span.hpp', make_regex(
    r'tcb::span'
    )),

    (r'tl/optional.hpp', make_regex(
    r'tl::optional'
    )),

)

# yapf: enable


def detect_includes(text: str) -> list[str]:
    out = []
    if text is not None:
        for header, regex in HEADER_MATCHERS:
            if regex.search(text) is not None:
                out.append(header)
    return out


IMPLICIT_INCLUDES = make_regex(
    r'version',
    r'cstdint',
    r'cstddef',
    r'cstdlib',
    r'numeric',
    r'type_traits',
    r'new',
    r'utility',
    r'memory',
    r'stdexcept',
    r'iterator',
    r'soagen.hpp',
)


def remove_implicit_includes(includes: list[str]) -> list[str]:
    return [inc for inc in includes if not IMPLICIT_INCLUDES.fullmatch(inc)]


RESERVED_CPP_KEYWORDS = make_regex(
    r'alignas',
    r'alignof',
    r'and',
    r'and_eq',
    r'asm',
    r'atomic_cancel',
    r'atomic_commit',
    r'atomic_noexcept',
    r'auto',
    r'bitand',
    r'bitor',
    r'bool',
    r'break',
    r'case',
    r'catch',
    r'char',
    r'char8_t',
    r'char16_t',
    r'char32_t',
    r'class',
    r'compl',
    r'concept',
    r'const',
    r'consteval',
    r'constexpr',
    r'constinit',
    r'const_cast',
    r'continue',
    r'co_await',
    r'co_return',
    r'co_yield',
    r'decltype',
    r'default',
    r'delete',
    r'do',
    r'double',
    r'dynamic_cast',
    r'else',
    r'enum',
    r'explicit',
    r'export',
    r'extern',
    r'false',
    r'float',
    r'for',
    r'friend',
    r'goto',
    r'if',
    r'inline',
    r'int',
    r'long',
    r'mutable',
    r'namespace',
    r'new',
    r'noexcept',
    r'not',
    r'not_eq',
    r'nullptr',
    r'operator',
    r'or',
    r'or_eq',
    r'private',
    r'protected',
    r'public',
    r'reflexpr',
    r'register',
    r'reinterpret_cast',
    r'requires',
    r'return',
    r'short',
    r'signed',
    r'sizeof',
    r'static',
    r'static_assert',
    r'static_cast',
    r'struct',
    r'switch',
    r'template',
    r'this',
    r'thread_local',
    r'throw',
    r'true',
    r'try',
    r'typedef',
    r'typeid',
    r'typename',
    r'union',
    r'unsigned',
    r'using',
    r'virtual',
    r'void',
    r'volatile',
    r'wchar_t',
    r'while',
    r'xor',
    r'xor_eq',
)

RESERVED_SOAGEN = make_regex(
    # std::vector-like interface:
    r'allocator_type',
    r'assign',
    r'at',
    r'begin',
    r'capacity',
    r'cbegin',
    r'cend',
    r'clear',
    r'data',
    r'difference_type',
    r'emplace_back',
    r'emplace',
    r'empty',
    r'end',
    r'erase',
    r'get_allocator',
    r'insert',
    r'max_size',
    r'push_back',
    r'reserve',
    r'resize',
    r'shrink_to_fit',
    r'size_type',
    r'size',
    r'swap',
    r'unordered_erase',
    # table machinery:
    r'aligned_stride',
    r'allocation_size',
    r'allocator_traits',
    r'column_indices',
    r'column_name',
    r'column_traits',
    r'column_type',
    r'column',
    r'columns',
    r'emplacer',
    r'for_each_column',
    r'for_each',
    r'forward_type',
    r'get',
    r'param_type',
    r'row_type',
    r'row',
    r'rvalue_type',
    r'sfinae',
    r'table_',
    r'table_traits',
    r'table_type',
    r'value_type',
    # 'unnamed' columns:
    r'first',
    r'second',
    r'third',
    r'fourth',
    r'fifth',
    r'sixth',
    r'seventh',
    r'eighth',
    r'ninth',
    r'tenth',
    r'eleventh',
    r'twelfth',
    r'thirteenth',
    r'fourteenth',
    r'fifteenth',
    r'sixteenth',
    # future-proofing:
    r'column_span',
    r'const_span',
    r'copy',
    r'find',
    r'index_type',
    r'move',
    r'reset',
    r'row_view',
    r'size_bytes',
    r'span',
    r'swap_columns',
    r'update',
    r'zerofill_column',
    r'zerofill',
    # other
    r'soagen',
    r'std',
)

VALID_IDENTIFIER = re.compile(r'^[A-Za-z][A-Za-z_0-9]*$')


def is_valid_identifier(s: str) -> tuple[bool, str]:
    if not s:
        return (False, 'may not be blank')
    if s.startswith('_'):
        return (False, 'may not begin with an underscore')
    if s.find('__') != -1:
        return (False, 'may not contain double-underscores')
    if re.fullmatch(r'^[0-9].*$', s):
        return (False, 'may not begin with a digit')
    if not VALID_IDENTIFIER.fullmatch(s):
        return (False, 'may contain only a-z, A-Z, 0-9, and underscores')
    if RESERVED_CPP_KEYWORDS.fullmatch(s):
        return (False, 'may not be a C++ keyword')
    if RESERVED_SOAGEN.fullmatch(s):
        return (False, 'reserved by soagen')
    return (True,)


__all__ = [r'detect_includes', r'remove_implicit_includes', r'is_valid_identifier']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import re
from io import StringIO
from pathlib import Path

from . import log, utils

# -----------------------------------------------------------------------------------------------------------------------
# code file writer object
# -----------------------------------------------------------------------------------------------------------------------


class Writer(object):
    def __init__(self, path, meta=None, indent='\t', clang_format=False, doxygen=True, on_flush=None):
        assert isinstance(path, Path)
        assert isinstance(indent, str)
        self.__path = path.resolve()
        self.__buffer = StringIO()
        self.__indent = 0
        self.__indent_str = str(indent)
        self.__indent_width = len(utils.untabify(self.__indent_str))
        self.__meta = meta
        self.__clang_format = bool(clang_format)
        self.__doxygen = bool(doxygen)
        self.__cpp_access_level = None
        self.__on_flush = on_flush
        self.emit_access_levels = True

    def write(self, text=None, end='\n', reindent=True):
        if text is None:
            text = ''
        elif not isinstance(text, str):
            text = str(text)
        if end is None:
            end = ''
        if end or text:
            if text:
                text = utils.replace_metavar(r'current_indent', self.__indent_str * self.__indent, text)
            if text and reindent:
                text = utils.reindent(text, indent=self.__indent_str * self.__indent)
            text += end
            if self.__meta is not None:
                text = self.__meta(text)
            self.__buffer.write(text)
        return self

    def __call__(self, text=None, **kwargs):
        return self.write(text=text, **kwargs)

    @property
    def indent(self) -> int:
        return self.__indent

    @indent.setter
    def indent(self, value: int):
        self.__indent = int(value)
        assert self.__indent >= 0

    @property
    def indent_width(self) -> int:
        return self.__indent * self.__indent_width

    @property
    def indent_str(self) -> str:
        return self.__indent_str

    @property
    def clang_format(self) -> bool:
        return self.__clang_format

    @property
    def doxygen(self) -> bool:
        return self.__doxygen

    @property
    def cpp_access_level(self) -> str:
        return self.__cpp_access_level

    @cpp_access_level.setter
    def cpp_access_level(self, value: str):
        assert value is None or value in (r'public', r'protected', r'private')
        if self.__cpp_access_level == value:
            return
        if self.__cpp_access_level == r'private' and self.doxygen and self.emit_access_levels:
            self('/// @endcond')
        self.__cpp_access_level = value
        if self.__cpp_access_level is not None and self.emit_access_levels:
            self()
            self(rf'{self.__cpp_access_level}:')
            if self.__cpp_access_level == r'private' and self.doxygen:
                self('/// @cond')

    def flush(self):
        s = self.__buffer.getvalue().lstrip()
        # do any user modifications via the on-flush callback
        if self.__on_flush is not None:
            s = str(self.__on_flush(self, s))
        # minor cleanup
        s = re.sub(r'\n(?:[ \t]*\n[ \t]*)+\n', '\n\n', s)  # double blank lines
        s = re.sub(r'([^@][({,])\n\n', r'\1\n', s)  # blank lines following opening brackets or a comma
        s = re.sub(r'\n\n([ \t]*[})])', r'\n\1', s)  # blank lines preceeding closing brackets
        s = re.sub(
            r'([ \t\n](?:namespace|struct|class|union)[ \t]+[a-zA-z0-9_:]+[ \t]*\n)\n+([ \t]*[{])', r'\1\2', s
        )  # blank lines between a namespace/class name and opening bracket
        # clang-format
        if self.__clang_format:
            log.i(f"Running clang-format for {self.__path}")
            try:
                s = utils.clang_format(s, cwd=self.__path.parent)
            except Exception as ex:
                log.w(f"Clang-format failed: {ex}")
        # write to disk
        log.i(f"Writing {self.__path}")
        with open(self.__path, 'w', encoding='utf-8', newline='\n') as f:
            f.write(s)
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.flush()


# -----------------------------------------------------------------------------------------------------------------------
# logical blocks within a file
# -----------------------------------------------------------------------------------------------------------------------


class WriterBlockBase(object):
    def __init__(self, writer: Writer):
        assert isinstance(writer, Writer)
        self.__writer = writer

    @property
    def writer(self) -> Writer:
        return self.__writer


class Indent(WriterBlockBase):
    def __init__(self, writer: Writer, pre='', post=''):
        super().__init__(writer)
        self.__pre = pre
        self.__post = post

    def __enter__(self):
        if self.__pre:
            self.writer(self.__pre)
        self.writer.indent += 1
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.writer.indent -= 1
        if self.__post:
            self.writer(self.__post)


class Namespace(WriterBlockBase):
    def __init__(self, writer: Writer, ns):
        super().__init__(writer)
        assert ns is None or isinstance(ns, str)
        self.__ok = False
        if ns is None:
            self.__ok = True
            self.__namespace = ''
        elif ns:
            self.__ok = True
            self.__namespace = ns

    def __enter__(self):
        if self.__ok:
            self.writer(f'namespace {self.__namespace}')
            self.writer('{')
            self.writer.indent += 1
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.__ok:
            self.writer.indent -= 1
            self.writer('}')


class Block(WriterBlockBase):
    def __init__(self, writer: Writer, pre='', post=''):
        super().__init__(writer)
        self.__pre = pre
        self.__post = post

    def __enter__(self):
        self.writer(rf'{self.__pre}{{')
        self.writer.indent += 1
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.writer.indent -= 1
        self.writer(rf'}}{self.__post}')


class ClassDefinition(WriterBlockBase):
    def __init__(self, writer: Writer, name: str, base_classes=None, hidden_base_classes=None):
        super().__init__(writer)
        self.__name = name
        self.__prev_cpp_access_level = None
        self.__bases = []
        if base_classes is not None:
            self.__bases += [b for b in utils.coerce_collection(base_classes)]
        self.__hidden_bases = []
        if hidden_base_classes is not None:
            self.__hidden_bases += [b for b in utils.coerce_collection(hidden_base_classes)]

    def __enter__(self):
        self.__prev_cpp_access_level = self.writer.cpp_access_level
        self.writer.cpp_access_level = None
        self.writer(self.__name + (' //' if self.__bases or self.__hidden_bases else ''))
        if self.__bases and self.__hidden_bases:
            self.writer(
                ': '
                + f',\n{self.writer.indent_str}'.join(self.__bases)
                + f'\n{self.writer.indent_str}SOAGEN_HIDDEN(, '
                + f',\n{self.writer.indent_str}'.join(self.__hidden_bases)
                + ')'
            )
        elif self.__bases:
            self.writer(': ' + f',\n{self.writer.indent_str}'.join(self.__bases))
        elif self.__hidden_bases:
            self.writer('SOAGEN_HIDDEN_BASE(' + f',\n{self.writer.indent_str}'.join(self.__hidden_bases) + ')')
        self.writer('{')
        self.writer.indent += 1
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.writer.indent -= 1
        self.writer('};')
        self.writer.emit_access_levels = False
        self.writer.cpp_access_level = self.__prev_cpp_access_level
        self.writer.emit_access_levels = True


class Private(WriterBlockBase):
    def __enter__(self):
        self.writer.cpp_access_level = r'private'
        self.writer.indent += 1
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.writer.indent -= 1
        self.writer()


class Protected(WriterBlockBase):
    def __enter__(self):
        self.writer.cpp_access_level = r'protected'
        self.writer.indent += 1
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.writer.indent -= 1
        self.writer()


class Public(WriterBlockBase):
    def __enter__(self):
        self.writer.cpp_access_level = r'public'
        self.writer.indent += 1
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.writer.indent -= 1
        self.writer()


class DoxygenGroup(WriterBlockBase):
    def __init__(self, writer: Writer, key, name):
        super().__init__(writer)
        self.__key = key
        self.__name = name

    def __enter__(self):
        if self.writer.doxygen:
            self.writer(
                rf'''
            /// @addtogroup {self.__key} {self.__name}
            /// @{{
            '''
            )
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.writer.doxygen:
            self.writer(
                r'''
            /// @}
            '''
            )


class DoxygenMemberGroup(WriterBlockBase):
    def __init__(self, writer: Writer, name: str, brief='', details='', availability=''):
        super().__init__(writer)
        self.__name = str(name)
        self.__brief = str(brief)
        self.__details = str(details)
        self.__availability = str(availability)

    def __enter__(self):
        if self.writer.doxygen:
            self.writer()
            self.writer(rf'/// @name {self.__name}')
            if self.__brief:
                self.writer(rf'/// @brief {self.__brief}')
            if self.__details:
                self.writer(rf'/// @details {self.__details}')
            if self.__availability:
                self.writer(rf'/// @availability {self.__availability}')
            self.writer(r'/// @{')
            self.writer()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.writer.doxygen:
            self.writer(
                r'''
            /// @}
            '''
            )


class HiddenFromDoxygen(WriterBlockBase):
    def __enter__(self):
        if self.writer.doxygen:
            self.writer(
                rf'''
            /// @cond
            '''
            )
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.writer.doxygen:
            self.writer(
                r'''
            /// @endcond
            '''
            )


class PreprocessorRegion(WriterBlockBase):
    def __init__(self, writer: Writer, condition='1'):
        super().__init__(writer)
        self.__condition = str(condition)

    def __enter__(self):
        self.writer(
            rf'''#if {self.__condition}
        '''
        )
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.writer(
            r'''
        #endif'''
        )


class ClangFormatOff(WriterBlockBase):
    def __enter__(self):
        self.writer()
        self.writer('// clang-format off')
        self.writer()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.writer()
        self.writer('// clang-format on')
        self.writer()


__all__ = [
    r'WriterBlockBase',
    r'Writer',
    r'Indent',
    r'Namespace',
    r'Block',
    r'ClassDefinition',
    r'Private',
    r'Protected',
    r'Public',
    r'DoxygenGroup',
    r'DoxygenMemberGroup',
    r'HiddenFromDoxygen',
    r'PreprocessorRegion',
    r'ClangFormatOff',
]



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE.txt for the full license text.
# SPDX-License-Identifier: MIT

from pathlib import Path
from . import paths

VERSION = ()

with open(paths.VERSION_TXT, encoding='utf-8') as file:
    VERSION = tuple([int(v.strip()) for v in file.read().strip().split('.')])
    assert len(VERSION) == 3

VERSION_STRING = r'.'.join([str(v) for v in VERSION])

__all__ = [r'VERSION', r'VERSION_STRING']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

from xml.sax.saxutils import escape

from . import utils
from .column import *
from .configurable import Configurable
from .metavars import *
from .writer import *
from .version import *


class NatvisFile(Configurable):
    def __init__(self, config, structs):
        super().__init__(config)
        self.path = self.config.path.with_suffix('.natvis')
        self.structs = utils.coerce_collection(structs)

        self.meta = MetaVars()
        for prefix in (r'file_', r'file::'):
            self.meta.push(rf'{prefix}name', self.path.name)
            self.meta.push(rf'{prefix}path', str(self.path))

    def write(self, o: Writer):
        with MetaScope(self):
            o(
                rf'''
            <?xml version="1.0" encoding="utf-8"?>
            <!--
            // This file was generated by soagen v{VERSION_STRING} - do not modify it directly
            // https://marzer.github.io/soagen
            -->
            <AutoVisualizer xmlns="http://schemas.microsoft.com/vstudio/debugger/natvis/2010">
            '''
            )

            with Indent(o):
                for struct in self.structs:
                    with MetaScope(struct):
                        o(
                            rf'''
                        <!--{"="*(120 - o.indent_width - 4)}
                        {struct.name}
                        {"="*(120 - o.indent_width - 3)}-->'''
                        )
                        with Indent(o, pre=f'\n\n<Type Name="{escape(struct.qualified_type)}">', post='</Type>'):
                            o(
                                r'''
                            <Intrinsic Name="size" Expression="table_.count_" />
                            <Intrinsic Name="size_bytes" Expression="table_.alloc_.size_in_bytes" />
                            <Intrinsic Name="capacity" Expression="table_.capacity_.first_" />
                            '''
                            )
                            for col in struct.columns:
                                o(
                                    rf'''
                                <Intrinsic
                                    Name="get_{col.index}"
                                    Expression="reinterpret_cast&lt;{escape(col.type)}*&gt;(table_.alloc_.columns[{col.index}])"
                                />
                                '''
                                )
                            o(r'<DisplayString>{{ size={size()} }}</DisplayString>')
                            with Indent(o, pre='<Expand>', post='</Expand>\n'):
                                o(
                                    r'''
                                <Item Name="[size]">size()</Item>
                                <Item Name="[capacity]">capacity()</Item>
                                <Item Name="[allocation_size]">size_bytes()</Item>
                                '''
                                )
                                for col in struct.columns:
                                    get = rf'get_{col.index}()'
                                    o(
                                        rf'''
                                    <Synthetic Name="{col.name}">
                                        <DisplayString Condition="size() &gt; 3u">{{{{ {{*({get})}}, {{*({get} + 1)}}, {{*({get} + 2)}}, ... }}}}</DisplayString>
                                        <DisplayString Condition="size() == 3u">{{{{ {{*({get})}}, {{*({get} + 1)}}, {{*({get} + 2)}} }}}}</DisplayString>
                                        <DisplayString Condition="size() == 2u">{{{{ {{*({get})}}, {{*({get} + 1)}} }}}}</DisplayString>
                                        <DisplayString Condition="size() == 1u">{{{{ {{*({get})}} }}}}</DisplayString>
                                        <DisplayString Condition="size() == 0u"></DisplayString>
                                        <Expand>
                                            <ArrayItems>
                                                <Size>size()</Size>
                                                <ValuePointer>{get}</ValuePointer>
                                            </ArrayItems>
                                        </Expand>
                                    </Synthetic>
                                    '''
                                    )

                        row_types = [
                            rf'soagen::row<{struct.qualified_type}, {struct.column_indices}>',
                            rf'soagen::row<{struct.qualified_type}&, {struct.column_indices}>',
                            rf'soagen::row<{struct.qualified_type}&&, {struct.column_indices}>',
                            rf'soagen::row<const {struct.qualified_type}, {struct.column_indices}>',
                            rf'soagen::row<const {struct.qualified_type}&, {struct.column_indices}>',
                            rf'soagen::row<const {struct.qualified_type}&&, {struct.column_indices}>',
                        ]
                        with Indent(o, pre=f'\n\n<Type Name="{escape(row_types[0])}">', post='</Type>'):
                            for i in range(1, len(row_types)):
                                o(rf'<AlternativeType Name="{escape(row_types[i])}" />')

                            s = r''
                            if len(struct.columns) > 1:
                                s += r'{{ '
                            s += r'{'
                            s += r'}, {'.join(col.name for col in struct.columns[:8])
                            s += r'}'
                            if len(struct.columns) > 8:
                                s += r', ...'
                            if len(struct.columns) > 1:
                                s += r' }}'
                            o(rf'<DisplayString>{s}</DisplayString>')
                            with Indent(o, pre='<Expand>', post='</Expand>\n'):
                                for col in struct.columns:
                                    o(rf'<Item Name="{col.name}">{col.name}</Item>')

            o(r'</AutoVisualizer>')


__all__ = [r'NatvisFile']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

"""
Constants for various key paths.
"""

import tempfile
from pathlib import Path

PACKAGE = Path(Path(__file__).resolve().parent)
"""The root directory of the package installation."""

SRC = Path(PACKAGE, r'..').resolve()
"""The root directory of repository's package sources."""

REPOSITORY = Path(SRC, r'..').resolve()
"""The root directory of the repository."""

TESTS = Path(REPOSITORY, r'tests')
"""The root directory of the repository's tests."""

HPP = Path(PACKAGE, r'hpp')
"""The PACKAGE containing soagen's (non-autogenerated) hpp files."""

HPP_TEMPLATES = Path(HPP, r'templates')
"""The directory containing soagen's .hpp template files"""

HPP_GENERATED = Path(HPP, r'generated')
"""The directory containing soagen's procedurally-generated .hpp files"""

HPP_SINGLE = Path(HPP, r'single')
"""The directory containing soagen's pre-processed 'single' soagen.hpp"""

DOCS = Path(REPOSITORY, r'docs')
"""The root directory of the repository's documentation."""

EXAMPLES = Path(REPOSITORY, r'examples')
"""The root directory of the repository's examples."""

TEMP = Path(tempfile.gettempdir(), r'soagen')
"""A global temp directory shared by all instances of soagen."""

BUG_REPORT_DIR = (TEMP / r'bug_report').resolve()
"""Directory used for assembling bug reports."""

BUG_REPORT_INPUTS = BUG_REPORT_DIR / r'inputs'
"""Directory used for storing the inputs when assembling bug reports."""

BUG_REPORT_OUTPUTS = BUG_REPORT_DIR / r'outputs'
"""Directory used for storing the outputs when assembling bug reports."""

VERSION_TXT = PACKAGE / r'version.txt'
"""The version file, version.txt."""


MAKE_SINGLE = Path(REPOSITORY, r'..', r'muu', r'tools', r'make_single.py')



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import argparse
import logging
import os
import re
import subprocess
import sys
import zipfile
from io import StringIO
from pathlib import Path

from . import cpp, log, paths, utils
from .config import Config
from .errors import Error
from .preprocessor import Preprocessor
from .schemas import SchemaError, current_schema_context
from .version import *
from .writer import Writer


def launch_worker(*args, cwd=None):
    if cwd is None:
        cwd = Path.cwd()
    cwd = utils.coerce_path(cwd)
    utils.run_python_script(
        Path(__file__).parent.parent / r'__main__.py',
        *args,
        r'--worker',
        r'--color' if log.styles_allowed() else r'--no-color',
        cwd=cwd,
    )


def bug_report():
    BUG_REPORT_STRIP_ARGS = (
        r'--bug-report',
        r'--bug-report-internal',
        r'-v',
        r'--verbose',
        r'--color',
        r'--no-color',
        r'--colour',
        r'--no-colour',
    )

    bug_report_args = [arg for arg in sys.argv[1:] if arg not in BUG_REPORT_STRIP_ARGS]
    bug_report_zip = (Path.cwd() / r'soagen_bug_report.zip').resolve()

    log.i(rf'{log.STYLE_CYAN}Preparing a bug report!{log.STYLE_RESET}')
    log.i(r'Preparing output paths')
    utils.delete_directory(paths.BUG_REPORT_DIR, logger=log.d)
    utils.delete_file(bug_report_zip, logger=log.d)
    os.makedirs(str(paths.BUG_REPORT_DIR), exist_ok=True)

    log.i(r'Invoking soagen')
    result = subprocess.run(
        args=[r'soagen', *bug_report_args, r'--bug-report-internal', r'--verbose', r'--no-color'],
        cwd=str(Path.cwd()),
        check=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding=r'utf-8',
    )

    if result.stdout is not None:
        log.i(r'Writing stdout')
        with open(paths.BUG_REPORT_DIR / r'stdout.txt', r'w', newline='\n', encoding=r'utf-8') as f:
            f.write(result.stdout)

    if result.stderr is not None:
        log.i(r'Writing stderr')
        with open(paths.BUG_REPORT_DIR / r'stderr.txt', r'w', newline='\n', encoding=r'utf-8') as f:
            f.write(result.stderr)

    log.i(r'Writing metadata')
    with open(paths.BUG_REPORT_DIR / r'metadata.txt', r'w', newline='\n', encoding=r'utf-8') as f:
        f.write(f'version: {VERSION_STRING}\n')
        f.write(f'args: {bug_report_args}\n')
        f.write(f'returncode: {result.returncode}\n')

    # zip file
    log.i(r'Zipping files')
    with zipfile.ZipFile(str(bug_report_zip), 'w', compression=zipfile.ZIP_DEFLATED, compresslevel=9) as zip:
        file_prefix_len = len(str(paths.BUG_REPORT_DIR))
        for file in utils.get_all_files(paths.BUG_REPORT_DIR, recursive=True):
            if file.suffix is not None and file.suffix.lower() in (r'.pyc',):
                continue
            relative_file = str(file)[file_prefix_len:].replace('\\', '/').strip('/')
            zip.write(file, arcname=rf'soagen_bug_report/{relative_file}')

    log.i(r'Cleaning up')
    utils.delete_directory(paths.BUG_REPORT_DIR)

    log.i(
        f'{log.STYLE_CYAN}Zip generated: {bug_report_zip}\n'
        f'Please attach this file when you make a report at github.com/marzer/soagen/issues, thanks!{log.STYLE_RESET}'
    )


def update(
    new_version: tuple = None,
    update_templated_hpps=True,
    update_soagen_hpp=True,
    update_examples=True,
    update_tests=True,
    update_documentation=True,
):
    # set the version number
    if new_version is not None:
        assert isinstance(new_version, tuple)
        assert len(new_version) == 3
        for val in new_version:
            assert isinstance(val, int) and val >= 0
        global VERSION
        global VERSION_STRING
        if VERSION != new_version:
            new_version_str = rf'{new_version[0]}.{new_version[1]}.{new_version[2]}'
            log.i(rf'Changing version number: v{VERSION_STRING} -> v{new_version_str}')
            matcher = re.compile(rf'\b({VERSION[0]})[.]({VERSION[1]})[.]({VERSION[2]})\b')
            matcher_v = re.compile(rf'\b([vV])({VERSION[0]})[.]({VERSION[1]})[.]({VERSION[2]})\b')

            # update version
            VERSION = new_version
            VERSION_STRING = new_version_str
            assert VERSION == new_version
            assert VERSION_STRING == new_version_str
            log.i(rf'Writing {paths.VERSION_TXT}')
            with open(paths.VERSION_TXT, 'w', encoding='utf-8', newline='\n') as f:
                print(new_version_str, file=f)

            # regenerate version.hpp
            version_hpp = paths.HPP_GENERATED / r'version.hpp'
            log.i(rf'Writing {version_hpp}')
            with open(version_hpp, 'w', encoding='utf-8', newline='\n') as f:
                f.write(
                    rf'''
//# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
//# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
//# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
//# SPDX-License-Identifier: MIT
#pragma once

#define SOAGEN_VERSION_MAJOR {VERSION[0]}
#define SOAGEN_VERSION_MINOR {VERSION[1]}
#define SOAGEN_VERSION_PATCH {VERSION[2]}
#define SOAGEN_VERSION_STRING "{VERSION_STRING}"
'''.lstrip()
                )

            # update version in all the other files
            DIRS = (
                (paths.REPOSITORY, False),
                (paths.DOCS, True),
                (paths.EXAMPLES, True),
                (paths.SRC, True),
                (paths.TESTS, True),
            )
            for dir, recursive in DIRS:
                for file in utils.enumerate_files(
                    dir,
                    any=(
                        r'*.html',
                        r'*.py',
                        r'*.txt',
                        r'*.md',
                        r'*.hpp',
                        r'*.cpp',
                        r'*.dox',
                        r'*.toml',
                        r'*.natvis',
                        r'meson.build',
                    ),
                    recursive=recursive,
                ):
                    if str(file).find('.egg-info') != -1 or file.name.lower().find('changelog') != -1:
                        continue
                    text = utils.read_all_text_from_file(file, logger=log.d)
                    new_text = matcher.sub(VERSION_STRING, text)
                    new_text = matcher_v.sub(lambda m: m[1] + VERSION_STRING, new_text)
                    if text != new_text:
                        log.i(rf'Writing updated version numbers in {file}')
                        with open(file, 'w', encoding='utf-8', newline='\n') as f:
                            f.write(new_text)

    # copy .clang-format
    if update_templated_hpps or update_soagen_hpp or update_tests or update_examples:
        utils.copy_file(Path(paths.REPOSITORY, r'.clang-format'), paths.HPP, logger=log.i)

    # regenerate the generated headers from their templates
    if update_templated_hpps and paths.MAKE_SINGLE.exists():
        for template in utils.enumerate_files(paths.HPP_TEMPLATES, all='*.hpp.in', recursive=True):
            output = Path(paths.HPP_GENERATED, template.stem)
            os.makedirs(str(paths.HPP_GENERATED), exist_ok=True)
            utils.run_python_script(
                *(
                    [
                        paths.MAKE_SINGLE,
                        str(template),
                        r'--output',
                        output,
                        r'--namespaces',
                        r'soagen',
                        r'detail',
                        r'--macros',
                        r'SOAGEN',
                    ]
                    + ([r'--no-strip-hidden-bases'] if template.name not in (r'compressed_pair.hpp.in',) else [])
                )
            )
            text = utils.read_all_text_from_file(output, logger=log.i)
            try:
                text = utils.clang_format(text, cwd=output.parent)
            except:
                pass
            log.i(rf'Writing {output}')
            with open(output, 'w', encoding='utf-8', newline='\n') as f:
                f.write(text)

    # read soagen.hpp + preprocess into single header
    if update_soagen_hpp:
        soagen_hpp_in = Path(paths.HPP, 'soagen.hpp')
        soagen_hpp_out = Path(paths.HPP_SINGLE, 'soagen.hpp')
        text = str(Preprocessor(soagen_hpp_in))
        text = utils.replace_metavar(r'version', VERSION_STRING, text)
        license = utils.read_all_text_from_file(paths.REPOSITORY / r'LICENSE.txt', log.i).replace('\r\n', '\n').strip()
        license = '\n// '.join(utils.reflow_text(license, line_length=117).split('\n'))
        text = utils.replace_metavar(r'license', license, text)
        log.i(rf'Writing {soagen_hpp_out}')
        os.makedirs(str(paths.HPP_SINGLE), exist_ok=True)
        with open(soagen_hpp_out, 'w', encoding='utf-8', newline='\n') as f:
            f.write(text)

    # re-run soagen itself on the examples and tests
    if update_examples and paths.EXAMPLES.exists():
        launch_worker(r'*.toml', r'--doxygen', cwd=paths.EXAMPLES)
    if update_tests and paths.TESTS.exists():
        launch_worker(r'*.toml', r'--no-doxygen', cwd=paths.TESTS)

    # invoke poxy on the documentation
    if update_documentation and utils.is_tool(r'poxy') and paths.DOCS.is_dir():
        subprocess.run(args=[r'poxy'], cwd=str(paths.DOCS), encoding=r'utf-8')


class NonExitingArgParser(argparse.ArgumentParser):
    def error(self, message):
        self.print_usage(sys.stderr)
        raise Error(message)


def main_impl():
    # yapf: disable
    args = NonExitingArgParser(
        description=
        rf'{log.STYLE_CYAN}'
        r'''  ___  ___   __ _  __ _  ___ _ __  ''' '\n'
        r''' / __|/ _ \ / _` |/ _` |/ _ \ '_ \ ''' '\n'
        r''' \__ \ (_) | (_| | (_| |  __/ | | |''' '\n'
        r''' |___/\___/ \__,_|\__, |\___|_| |_|''' '\n'
        r'''                   __/ |           ''' '\n'
        r'''                  |___/  '''  rf'{log.STYLE_RESET}'
        rf' v{VERSION_STRING} - github.com/marzer/soagen'
        '\n\n'
        r'Struct-of-Arrays generator for C++ projects.',
        formatter_class=argparse.RawTextHelpFormatter,
        exit_on_error=False
    )
    # yapf: enable

    # --------------------------------------------------------------
    # public (user-facing, documented) arguments
    # --------------------------------------------------------------

    args.add_argument(
        r'configs',
        type=str,
        nargs=r'*',
        help="zero or more .toml files describing your structures-of-arrays\n(wildcards are accepted, e.g. soa/*.toml)",
    )
    args.add_argument(r'-v', r'--verbose', action=r'store_true', help=r"enable very noisy diagnostic output")
    args.add_argument(r'--version', action=r'store_true', help=r"print the version and exit", dest=r'print_version')
    args.add_argument(
        r'--install',  #
        type=Path,
        default=None,
        metavar=r'<dir>',
        help=rf"install {log.STYLE_CYAN}soagen.hpp{log.STYLE_RESET} into a directory",
    )
    args.add_argument(
        r'--werror',  #
        action=argparse.BooleanOptionalAction,
        default=False,
        help=f'treat {log.STYLE_YELLOW}warnings{log.STYLE_RESET} as {log.STYLE_RED}errors{log.STYLE_RESET} (default: %(default)s)',
    )
    args.add_argument(
        r'--color',  #
        action=argparse.BooleanOptionalAction,
        default=None,
        help=f'use {log.STYLE_RED}c{log.STYLE_YELLOW}o{log.STYLE_GREEN}l{log.STYLE_BLUE}o'
        + f'{log.STYLE_MAGENTA}r{log.STYLE_CYAN}s{log.STYLE_RESET} in terminal output (default: True)\n'
        r'(the British spelling "colour" is also accepted)',
    )
    args.add_argument(
        r'--clang-format',  #
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f'attempt to run {log.STYLE_CYAN}clang-format{log.STYLE_RESET} on generated code (default: %(default)s)',
    )
    args.add_argument(
        r'--doxygen',  #
        action=argparse.BooleanOptionalAction,
        default=False,
        help=f'include {log.STYLE_CYAN}doxygen{log.STYLE_RESET} markup in the generated code (default: %(default)s)',
    )
    args.add_argument(
        r'--natvis',  #
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f'generate {log.STYLE_CYAN}.natvis{log.STYLE_RESET} files for Visual Studio (default: %(default)s)',
    )
    args.add_argument(
        r'--bug-report', action=r'store_true', help=r"capture all inputs and outputs in a bug-report zip file"
    )  #

    # --------------------------------------------------------------
    # hidden/developer-only/deprecated/diagnostic arguments
    # --------------------------------------------------------------

    args.add_argument(r'--where', action=r'store_true', help=argparse.SUPPRESS)
    args.add_argument(r'--update', nargs='?', const=True, default=False, help=argparse.SUPPRESS)
    args.add_argument(r'--update-hpp', action=r'store_true', help=argparse.SUPPRESS)
    args.add_argument(r'--update-docs', action=r'store_true', help=argparse.SUPPRESS)
    args.add_argument(r'--bug-report-internal', action=r'store_true', help=argparse.SUPPRESS)
    args.add_argument(
        r'--colour', action=argparse.BooleanOptionalAction, default=None, dest='color', help=argparse.SUPPRESS
    )
    args.add_argument(r'--worker', action=r'store_true', help=argparse.SUPPRESS)

    usage_str = args.format_usage()
    args = args.parse_args()

    log.reinit(
        min_level=logging.DEBUG if args.verbose else logging.INFO,
        treat_warnings_as_errors=args.werror,
        on_warning=(lambda: sys.exit(1)) if args.werror else None,
        on_error=lambda: sys.exit(1),
    )

    # --------------------------------------------------------------
    # --version
    # --------------------------------------------------------------

    if args.print_version:
        print(VERSION_STRING)
        return

    # --------------------------------------------------------------
    # --where
    # --------------------------------------------------------------

    if args.where:
        print(paths.PACKAGE)
        return

    if not args.worker:
        log.i(rf'{log.STYLE_CYAN}soagen{log.STYLE_RESET} v{VERSION_STRING}')

    # --------------------------------------------------------------
    # bug report invocation
    # --------------------------------------------------------------

    if args.bug_report:
        bug_report()
        return

    # --------------------------------------------------------------
    # regular invocation
    # --------------------------------------------------------------

    done_work = False

    if args.bug_report_internal:
        os.makedirs(str(paths.BUG_REPORT_OUTPUTS), exist_ok=True)

        # bug reports should always copy the submitter's version of
        # soagen.hpp in case the report is related to soagen.hpp
        utils.copy_file(paths.HPP / r'soagen.hpp', paths.BUG_REPORT_DIR, logger=log.i)

    if args.update:
        new_version = None
        if isinstance(args.update, str):
            args.update = args.update.strip()
            if args.update:
                m = re.fullmatch(r'\s*[vV]?\s*([0-9]+)\s*[.]+\s*([0-9]+)\s*[.]+\s*([0-9]+)\s*', args.update)
                if m:
                    new_version = (int(m[1]), int(m[2]), int(m[3]))
        done_work = True
        update(new_version)
    elif args.update_hpp or args.update_docs:
        done_work = True
        update(
            update_templated_hpps=False,
            update_soagen_hpp=args.update_hpp,
            update_examples=False,
            update_tests=False,
            update_documentation=args.update_docs,
        )

    if args.install is not None:
        done_work = True
        args.install: Path
        if not args.install.exists() or not args.install.is_dir():
            log.e(rf"--install: path '{args.install}' did not exist or was not a directory")
        log.i(rf"Copying soagen.hpp to {args.install.resolve()}")
        utils.copy_file(paths.HPP_SINGLE / r'soagen.hpp', args.install)

    configs = set()
    for config in args.configs:
        if config == '.':
            config = '*.toml'
        if config.find('*') != -1:
            configs.update([f for f in Path('.').glob(config) if f.is_file()])
        else:
            config = Path(config)
            if config not in configs:
                if not config.exists() or not config.is_file():
                    log.e(rf"configs: '{config}' did not exist or was not a file")
                configs.add(config)
    configs = sorted(configs)
    if args.bug_report_internal:
        os.makedirs(str(paths.BUG_REPORT_INPUTS), exist_ok=True)
        for f in configs:
            utils.copy_file(f, paths.BUG_REPORT_INPUTS, log.i)
    configs = [Config(f) for f in configs]
    done_work = done_work or bool(configs)

    if configs:
        if args.clang_format and not utils.is_tool(r'clang-format'):
            log.w(r'clang-format not found on system PATH')
            args.clang_format = False

        for config in configs:
            for src in config.hpp:

                def on_flush(o: Writer, s: str) -> str:
                    nonlocal src
                    # sub in external headers
                    includes = sorted(set(src.includes.external + cpp.detect_includes(s)))
                    includes = cpp.remove_implicit_includes(includes)
                    EXTERNAL_HEADERS = r'\n[ \t]*//[ \t]*__SOAGEN_EXTERNAL_HEADERS[ \t]*\n'
                    rep = '\nSOAGEN_DISABLE_WARNINGS;'
                    for inc in includes:
                        rep += f'\n#include <{inc}>'
                    rep += '\nSOAGEN_ENABLE_WARNINGS;\n'
                    s = re.sub(EXTERNAL_HEADERS, rep, s)
                    # strip doxygen stuff if we have that disabled
                    if not o.doxygen:
                        s = re.sub(
                            r'\n[ \t]*#[ \t]*if[ \t]*SOAGEN_DOXYGEN[ \t]*\n.+?\n[ \t]*#[ \t]*endif[ \t]*\n',
                            '\n',
                            s,
                            flags=re.DOTALL,
                        )
                        s = re.sub(
                            r'SOAGEN_ENABLE_IF_T\s*\(\s*(.+?)\s*,\s*sfinae\s*\)',
                            r'std::enable_if_t<sfinae, \1>',
                            s,
                            flags=re.DOTALL,
                        )
                        s = re.sub(
                            r'SOAGEN_ENABLE_IF\s*\(\s*\(\s*(.+?)\s*\)\s*\)',  # two sets of brackets
                            r', std::enable_if_t<(\1), int> = 0',
                            s,
                            flags=re.DOTALL,
                        )
                        s = re.sub(  # one set of brackets
                            r'SOAGEN_ENABLE_IF\s*\(\s*(.+?)\s*\)',
                            r', std::enable_if_t<\1, int> = 0',
                            s,
                            flags=re.DOTALL,
                        )
                        s = re.sub(
                            r'SOAGEN_HIDDEN_PARAM\s*\(\s*bool\s*sfinae\s*=\s*\(\s*(.+?)\s*\)\s*\)',  # two sets of brackets
                            r', bool sfinae = (\1)',
                            s,
                            flags=re.DOTALL,
                        )
                        s = re.sub(r'SOAGEN_HIDDEN_BASE\s*\(\s*(.+?)\s*\)', r': \1', s, flags=re.DOTALL)
                        s = re.sub(
                            r'SOAGEN_HIDDEN\s*\(\s*template\s*<(.+?)\s*>\s*\)', r'template <\1>', s, flags=re.DOTALL
                        )
                    return s

                with Writer(
                    src.path,
                    meta=config.meta_stack,
                    clang_format=args.clang_format,
                    doxygen=args.doxygen,
                    on_flush=on_flush,
                ) as o:
                    src.write(o)

                if args.bug_report_internal:
                    utils.copy_file(src.path, paths.BUG_REPORT_OUTPUTS, logger=log.i)

            if args.natvis:
                for src in (config.natvis,):
                    with Writer(src.path, meta=config.meta_stack, clang_format=False, doxygen=False) as o:
                        src.write(o)

                if args.bug_report_internal:
                    utils.copy_file(src.path, paths.BUG_REPORT_OUTPUTS, logger=log.i)

    if not args.worker:
        if done_work:
            log.i(r'All done!')
        else:
            log.i(r'No work to do.')
            if len(sys.argv) == 1:
                log.i(usage_str)


def main():
    allow_styles = True
    for arg in (r'--no-color', r'--no-colour'):
        if arg in sys.argv[1:]:
            allow_styles = False
            break
    log.init(allow_styles=allow_styles)
    try:
        try:
            main_impl()
        except BaseException as err:
            log.clear_hooks()
            raise err from None
    except SystemExit as exit:
        raise exit from None
    except SchemaError as err:
        log.e(rf'{current_schema_context()}{err}')
        sys.exit(1)
    except (Error, SchemaError, argparse.ArgumentError) as err:
        log.e(err)
        sys.exit(1)
    except BaseException as err:
        with StringIO() as buf:
            buf.write(
                f'\n{log.STYLE_RED_DIM}*************{log.STYLE_RESET}\n\n'
                'You appear to have triggered an internal bug!'
                f'\n{log.STYLE_CYAN}Please re-run soagen with --bug-report and file an issue at github.com/marzer/soagen/issues{log.STYLE_RESET}'
                '\nMany thanks!'
                f'\n\n{log.STYLE_RED_DIM}*************{log.STYLE_RESET}\n\n'
            )
            utils.print_exception(err, include_type=True, include_traceback=True, skip_frames=1, logger=buf)
            buf.write(f'{log.STYLE_RED_DIM}*************{log.STYLE_RESET}\n')
            print(buf.getvalue(), file=sys.stderr)
        sys.exit(1)
    sys.exit(0)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import re
from pathlib import Path

from . import utils


def get_relative_path(path: Path, relative_to: Path) -> Path:
    relative_root = relative_to
    relative = None
    relative_prefix = ''
    while True:
        try:
            relative = path.relative_to(relative_root)
            break
        except:
            if not relative_root.is_dir() or relative_root.parent == relative_root:
                break
            relative_prefix = rf'../{relative_prefix}'
            relative_root = relative_root.parent
            pass
    if relative is not None:
        return Path(rf'{relative_prefix}{relative}')
    return path


RX_INCLUDES = re.compile(r'^\s*#\s*include\s+"(.+?)".*?$', flags=re.I | re.M)
RX_PRAGMA_ONCE = re.compile(r'^\s*#\s*pragma\s+once\s*$', flags=re.M)

# explicit 'strip this' blocks
RX_STRIP_BLOCKS = re.compile(r'(?:\n[ \t]*)?//[#!][ \t]*[{][{].*?//[#!][ \t]*[}][}].*?\n', flags=re.S)

header_separator = r'(?://[ \t]*\*\*\*\**[ \t]*[a-zA-Z0-9_/.-]+[ \t]*\*\*\*\*+\n)'

# consecutive header end/header start
RX_HEADER_END_START_OUTER = re.compile(
    r'//[ \t]*__SOAGEN_HEADER_END\{\{.+?\}\}__SOAGEN_HEADER_START[ \t]*\n', flags=re.DOTALL
)
RX_HEADER_END_START_INNER = re.compile(
    rf'//[ \t]*\}}\}}__SOAGEN_HEADER_END\n+({header_separator})\n+//[ \t]*__SOAGEN_HEADER_START\{{\{{[ \t]*\n',
    flags=re.DOTALL,
)
RX_HEADER_START_END = re.compile(r'//[ \t]*(?:\}\}[ \t]*)?__SOAGEN_HEADER_(?:START|END)[ \t]*(?:\{\{[ \t]*)?\n')

# misc cleanup
RX_CLEANUP = (
    # trailing whitespace
    (re.compile(r'([^ \t])[ \t]+\n'), r'\1\n'),
    # double blank lines
    (re.compile('\n\n\n+'), '\n\n'),
    # magic comments
    (re.compile(rf'(?://(?:[#!</]|[ \t]?(?:\^\^\^|vvv))).*?($)', flags=re.MULTILINE), r'\1'),
    # consecutive header separators
    (re.compile(rf'(?:{header_separator}(?:[ \t]*\n)*)+({header_separator})'), r'\1'),
    # weird spacing edge case around pp directives
    (re.compile('\n[}]\n#'), r'\n}\n\n#'),
    # blank lines following a comma, opening bracket, or colon
    (re.compile(r'([{(\[:,])\n\n'), r'\1\n'),
    # blank lines preceeding closing brackets
    (re.compile(r'\n\n([ \t]*[\]})])'), r'\n\1'),
    # empty #if blocks
    (re.compile(r'#if(n?def)?[ \t]+[a-zA-Z0-9_]+\s*?\n+\s*?#endif'), r'\n'),
    # empty #else blocks
    (re.compile(r'#else\s*?\n+\s*?#endif'), r'#endif\n'),
    # #if SOAGEN_DOXYGEN ... #else ... #endif
    (re.compile(r'#if[ \t]+SOAGEN_DOXYGEN\s*?(?:\n+[^#]*?)?\n+\s*?#else\s*?\n+([^#]*?)\n+\s*?#endif'), r'\1\n'),
    # #if !SOAGEN_DOXYGEN ... #else ... #endif
    (re.compile(r'#if[ \t]+![ \t]*SOAGEN_DOXYGEN\s*?\n+([^#]*?)\n+\s*?#else\s*?(?:\n+[^#]*?)?\n+\s*?#endif'), r'\1\n'),
    # #if SOAGEN_DOXYGEN ... #endif
    (re.compile(r'#if[ \t]+SOAGEN_DOXYGEN\s*?(?:\n+[^#]*?)?\n+\s*?#endif'), r'\n'),
    # #if !SOAGEN_DOXYGEN ... #endif
    (re.compile(r'#if[ \t]+![ \t]*SOAGEN_DOXYGEN\s*?\n+([^#]*?)\n+\s*?#endif'), r'\1\n'),
)


class Preprocessor:

    """
    A simple C++ file 'preprocessor' for almalgamating source files.
    """

    __repeatable_headers = (r'header_start.hpp', r'header_end.hpp')

    def __init__(self, file):
        self.__processed_files = set()
        self.__once_only = set()
        self.__directory_stack = [Path.cwd()]
        self.__include_stack = []
        self.__entry_root = None
        self.__string = self.__preprocess(file)
        # do some cleanup of the preprocessed file
        while True:
            s = self.__string
            # misc cleanup
            for rx, repl in RX_CLEANUP:
                s = rx.sub(repl, s)
            # remove bookended header_end / header_start
            pos = 0
            m = RX_HEADER_END_START_OUTER.search(s)
            while m:
                m2 = RX_HEADER_END_START_INNER.search(m[0])
                if m2:
                    new_s = s[: m.start()] + m2[1]
                    pos = len(new_s)
                    new_s += s[m.end() :]
                    s = new_s
                else:
                    pos = m.end()
                m = RX_HEADER_END_START_OUTER.search(s, pos=pos)
            s = RX_HEADER_START_END.sub('\n', s)
            if s == self.__string:
                break
            self.__string = s
        self.__string = self.__string.strip() + '\n'

    def __preprocess(self, incl):
        if not isinstance(incl, (Path, str)):  # a regex match object
            incl = incl.group(1).strip()
        if isinstance(incl, str):
            incl = Path(incl.strip().replace('\\', r'/'))
        if not incl.is_absolute():
            incl = Path(self.__directory_stack[-1], incl).resolve()
        if not incl.is_file():
            sibling = Path(incl.parent.parent, 'hpp', incl.name)
            if sibling.is_file():
                incl = sibling
        self.__processed_files.add(incl)
        if incl in self.__once_only:
            return ''
        self.__include_stack.append(incl)

        text = utils.read_all_text_from_file(incl, logger=True).strip() + '\n'
        text = text.replace('\r\n', '\n')  # convert windows newlines

        self.__directory_stack.append(incl.parent)
        if RX_PRAGMA_ONCE.search(text):
            self.__once_only.add(incl)
        if len(self.__include_stack) == 1 and self.__entry_root is None:
            self.__entry_root = incl.parent.resolve()
        if len(self.__include_stack) > 1:
            text = RX_PRAGMA_ONCE.sub('', text)

        text = RX_INCLUDES.sub(lambda m: self.__preprocess(m), text, 0)
        text = RX_STRIP_BLOCKS.sub('\n', text)

        incl_name = incl.name
        incl = get_relative_path(incl, self.__entry_root)
        incl = str(incl).replace('\\', r'/').strip('/')

        if len(self.__include_stack) > 1 and incl_name not in self.__repeatable_headers:
            header = utils.make_divider(str(incl), 10, pattern=r'*')

            footer = ''
            if len(self.__include_stack) > 2:
                footer = get_relative_path(self.__include_stack[-2], self.__entry_root)
                footer = str(footer).replace('\\', r'/').strip('/')
                footer = utils.make_divider(footer, 10, pattern=r'*')

            text = f'\n\n{header}\n\n{text}\n\n{footer}'.rstrip()

        self.__include_stack.pop()
        self.__directory_stack.pop()
        return '\n\n' + text + '\n\n'

    def __str__(self):
        return self.__string

    def processed_files(self):
        out = list(self.__processed_files)
        out.sort()
        return out


__all__ = [r'Preprocessor']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

from . import cpp
from .configurable import *
from .schemas import *


class Includes(Configurable):
    __schema = Schema(
        {
            Optional(r'internal', default=list): And(
                ValueOrArray(str, name=r'internal'),
                Use(lambda x: cpp.remove_implicit_includes(sorted(set([s.strip() for s in x if s])))),
            ),
            Optional(r'external', default=list): And(
                ValueOrArray(str, name=r'external'),
                Use(lambda x: cpp.remove_implicit_includes(sorted(set([s.strip() for s in x if s])))),
            ),
        }
    )

    def __init__(self, cfg, vals):
        super().__init__(cfg)

        vals = Includes.__schema.validate(vals)
        self.__dict__.update(vals)


__all__ = [r'Includes']
